import os
import datetime
import numpy as np
import tensorflow as tf
import scipy.sparse as sp

from tqdm import tqdm
from tensorflow.keras import backend as K
from tensorflow.keras.utils import Sequence
from tensorflow.python.keras import callbacks as callbacks_module
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint

from graphgallery.nn.models import BaseModel
from graphgallery.utils.history import History
from graphgallery import asintarr, astensor


class SupervisedModel(BaseModel):
    """
        Base model for supervised learning.

        Arguments:
        ----------
            adj: shape (N, N), Scipy sparse matrix if  `is_adj_sparse=True`, 
                Numpy array-like (or matrix) if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            features: Numpy array-like with shape (N, F)
                The input node feature matrix, where `F` is the dimension of node features.
            labels: Numpy array-like with shape (N,)
                The ground-truth labels for all nodes in graph.
            device (String, optional): 
                The device where the model is running on. You can specified `CPU` or `GPU` 
                for the model. (default: :str: `CPU:0`, i.e., running on the 0-th `CPU`)
            seed (Positive integer, optional): 
                Used in combination with `tf.random.set_seed` & `np.random.seed` & `random.seed`  
                to create a reproducible sequence of tensors across multiple calls. 
                (default :obj: `None`, i.e., using random seed)
            name (String, optional): 
                Specified name for the model. (default: :str: `class.__name__`)

    """

    def __init__(self, adj, x, labels=None, device='CPU:0', seed=None, name=None, **kwargs):
        super().__init__(adj, x, labels, device, seed, name, **kwargs)

    def build(self):
        """
            Build the model using customized hyperparameters.

        Note:
        ----------
            This method must be called before training/testing/predicting. 
            Use `model.build()`. The following `Arguments` are only commonly used 
            arguments, and other model-specific arguments are not introduced as follows.


        Arguments:
        ----------
            hiddens: `list` of integer 
                The number of hidden units of model. Note: the last hidden unit (`n_classes`)
                aren't nececcary to specified and it will be automatically added in the last 
                layer. 
            activations: `list` of string
                The activation function of model. Note: the last activation function (`softmax`) 
                aren't nececcary to specified and it will be automatically spefified in the 
                final output.              
            dropout: Float scalar
                Dropout rate for the hidden outputs.
            lr: Float scalar
                Learning rate for the training model.
            l2_norm: Float scalar
                L2 normalize for the hidden layers. (usually only used in the first layer)
            use_bias: Boolean
                Whether to use bias in the hidden layers.

        """
        raise NotImplementedError

    def train(self, index_train, index_val=None,
              epochs=200, early_stopping=None,
              verbose=False, save_best=True, weight_path=None, as_model=False,
              monitor='val_acc', early_stop_metric='val_loss'):
        """
            Train the model for the input `index_train` of nodes or `sequence`.

        Note:
        ----------
            You must compile your model before training/testing/predicting. Use `model.build()`.

        Arguments:
        ----------
            index_train: Numpy array-like, `list`, Integer scalar or `graphgallery.NodeSequence`
                the index of nodes (or sequence) that will be used during training.    
            index_val: Numpy array-like, `list`, Integer scalar or `graphgallery.NodeSequence`, optional
                the index of nodes (or sequence) that will be used for validation. 
                (default :obj: `None`, i.e., do not use validation during training)
            epochs: Postive integer
                The number of epochs of training.(default :obj: `200`)
            early_stopping: Postive integer or None
                The number of early stopping patience during training. (default :obj: `None`, 
                i.e., do not use early stopping during training)
            verbose: Boolean
                Whether to show the training details. (default :obj: `None`)
            save_best: Boolean
                Whether to save the best weights (accuracy of loss depend on `monitor`) 
                of training or validation (depend on `validation` is `False` or `True`). 
                (default :bool: `True`)
            weight_path: String or None
                The path of saved weights/model. (default :obj: `None`, i.e., 
                `./log/{self.name}_weights`)
            as_model: Boolean
                Whether to save the whole model or weights only, if `True`, the `self.custom_objects`
                must be speficied if you are using customized `layer` or `loss` and so on.
            monitor: String
                One of (val_loss, val_acc, loss, acc), it determines which metric will be
                used for `save_best`. (default :obj: `val_acc`)
            early_stop_metric: String
                One of (val_loss, val_acc, loss, acc), it determines which metric will be 
                used for early stopping. (default :obj: `val_loss`)

        Return:
        ----------
            history: graphgallery.utils.History
                tensorflow like `history` instance.
        """
        # TODO use tensorflow callbacks

        # Check if model has been built
        if not self.model:
            raise RuntimeError('You must compile your model before training/testing/predicting. Use `model.build()`.')

        if isinstance(index_train, Sequence):
            train_data = index_train
        else:
            train_data = self.train_sequence(index_train)
            self.index_train = asintarr(index_train)

        validation = index_val is not None

        if validation:
            if isinstance(index_val, Sequence):
                val_data = index_val
            else:
                val_data = self.test_sequence(index_val)
                self.index_val = asintarr(index_val)

        history = History(monitor_metric=monitor,
                          early_stop_metric=early_stop_metric)

        if weight_path is None:
            weight_path = self.weight_path

        if not weight_path.endswith('.h5'):
            weight_path += '.h5'

        if not validation:
            history.register_monitor_metric('acc')
            history.register_early_stop_metric('loss')

        if verbose:
            pbar = tqdm(range(1, epochs+1))
        else:
            pbar = range(1, epochs+1)

        for epoch in pbar:

            if self.do_before_train is not None:
                self.do_before_train()

            loss, accuracy = self.do_forward(train_data)
            train_data.on_epoch_end()

            history.add_results(loss, 'loss')
            history.add_results(accuracy, 'acc')

            if validation:
                if self.do_before_validation is not None:
                    self.do_before_validation()

                val_loss, val_accuracy = self.do_forward(val_data, training=False)

                history.add_results(val_loss, 'val_loss')
                history.add_results(val_accuracy, 'val_acc')

            # record eoch and running times
            history.record_epoch(epoch)

            if save_best and history.save_best:
                self.save(weight_path, as_model=as_model)

            # early stopping
            if early_stopping and history.time_to_early_stopping(early_stopping):
                msg = f'Early stopping with patience {early_stopping}.'
                if verbose:
                    pbar.set_description(msg)
                    pbar.close()
                break

            if verbose:
                msg = f'loss {loss:.2f}, acc {accuracy:.2%}'
                if validation:
                    msg += f', val_loss {val_loss:.2f}, val_acc {val_accuracy:.2%}'
                pbar.set_description(msg)

        if save_best:
            self.load(weight_path, as_model=as_model)
            os.remove(weight_path)

        return history

    def train_v2(self, index_train, index_val=None,
                 epochs=200, early_stopping=None,
                 verbose=False, save_best=True, weight_path=None, as_model=False,
                 monitor='val_acc', early_stop_metric='val_loss', callbacks=None, **kwargs):
        """
            Train the model for the input `index_train` of nodes or `sequence`.

        Note:
        ----------
            You must compile your model before training/testing/predicting. Use `model.build()`.

        Arguments:
        ----------
            index_train: Numpy array-like, `list`, Integer scalar or `graphgallery.NodeSequence`
                the index of nodes (or sequence) that will be used during training.    
            index_val: Numpy array-like, `list`, Integer scalar or `graphgallery.NodeSequence`, optional
                the index of nodes (or sequence) that will be used for validation. 
                (default :obj: `None`, i.e., do not use validation during training)
            epochs: Postive integer
                The number of epochs of training.(default :obj: `200`)
            early_stopping: Postive integer or None
                The number of early stopping patience during training. (default :obj: `None`, 
                i.e., do not use early stopping during training)
            verbose: Boolean
                Whether to show the training details. (default :obj: `None`)
            save_best: Boolean
                Whether to save the best weights (accuracy of loss depend on `monitor`) 
                of training or validation (depend on `validation` is `False` or `True`). 
                (default :bool: `True`)
            weight_path: String or None
                The path of saved weights/model. (default :obj: `None`, i.e., 
                `./log/{self.name}_weights`)
            as_model: Boolean
                Whether to save the whole model or weights only, if `True`, the `self.custom_objects`
                must be speficied if you are using customized `layer` or `loss` and so on.
            monitor: String
                One of (val_loss, val_acc, loss, acc), it determines which metric will be
                used for `save_best`. (default :obj: `val_acc`)
            early_stop_metric: String
                One of (val_loss, val_acc, loss, acc), it determines which metric will be 
                used for early stopping. (default :obj: `val_loss`)
            callbacks: tensorflow.keras.callbacks. (default :obj: `None`)
            kwargs: other keyword arguments.

        Return:
        ----------
            A `tf.keras.callbacks.History` object. Its `History.history` attribute is
            a record of training loss values and metrics values
            at successive epochs, as well as validation loss values
            and validation metrics values (if applicable).

        """

        if not tf.__version__ >= '2.2.0':
            raise RuntimeError(f'This method is only work for tensorflow version >= 2.2.0.')

        # Check if model has been built
        if not self.model:
            raise RuntimeError('You must compile your model before training/testing/predicting. Use `model.build()`.')

        if isinstance(index_train, Sequence):
            train_data = index_train
        else:
            train_data = self.train_sequence(index_train)
            self.index_train = asintarr(index_train)

        validation = index_val is not None

        if validation:
            if isinstance(index_val, Sequence):
                val_data = index_val
            else:
                val_data = self.test_sequence(index_val)
                self.index_val = asintarr(index_val)

        model = self.model

        if not isinstance(callbacks, callbacks_module.CallbackList):
            callbacks = callbacks_module.CallbackList(callbacks,
                                                      add_history=True,
                                                      add_progbar=True,
                                                      verbose=verbose,
                                                      epochs=epochs)
        if early_stopping is not None:
            es_callback = EarlyStopping(monitor=early_stop_metric,
                                        patience=early_stopping,
                                        mode='auto',
                                        verbose=kwargs.pop('es_verbose', 0))
            callbacks.append(es_callback)

        if save_best:
            if weight_path is None:
                weight_path = self.weight_path

            if not weight_path.endswith('.h5'):
                weight_path += '.h5'

            mc_callback = ModelCheckpoint(weight_path,
                                          monitor=monitor,
                                          save_best_only=True,
                                          save_weights_only=not as_model,
                                          verbose=verbose)
            callbacks.append(mc_callback)
        callbacks.set_model(model)

        # leave it blank for the future
        allowed_kwargs = set([])
        unknown_kwargs = set(kwargs.keys()) - allowed_kwargs
        if unknown_kwargs:
            raise TypeError(
                "Invalid keyword argument(s) in `__init__`: %s" % (unknown_kwargs,))

        callbacks.on_train_begin()

        for epoch in range(epochs):
            callbacks.on_epoch_begin(epoch)

            if self.do_before_train is not None:
                self.do_before_train()

            callbacks.on_train_batch_begin(0)
            loss, accuracy = self.do_forward(train_data)
            train_data.on_epoch_end()

            training_logs = {'loss': loss, 'acc': accuracy}
            callbacks.on_train_batch_end(0, training_logs)

            if validation:
                if self.do_before_validation is not None:
                    self.do_before_validation()

                val_loss, val_accuracy = self.do_forward(val_data, training=False)
                training_logs.update({'val_loss': val_loss, 'val_acc': val_accuracy})

            callbacks.on_epoch_end(epoch, training_logs)

            if model.stop_training:
                break

        callbacks.on_train_end()

        if save_best:
            self.load(weight_path, as_model=as_model)
            if os.exist(weight_path):
                os.remove(weight_path)

        return model.history

    def test(self, index, **kwargs):
        """
            Test the output accuracy for the `index` of nodes or `sequence`.

        Note:
        ----------
            You must compile your model before training/testing/predicting.
            Use `model.build()`.

        Arguments:
        ----------
            index: Numpy array-like, `list`, Integer scalar or `graphgallery.NodeSequence`
                The index of nodes (or sequence) that will be tested.    

            **kwargs (optional): Additional arguments of
                :method:`do_before_test`.   

        Return:
        ----------
            loss: Float scalar
                Output loss of forward propagation. 
            accuracy: Float scalar
                Output accuracy of prediction.        
        """

        # TODO record test logs like self.train() or self.train_v2()
        if not self.model:
            raise RuntimeError('You must compile your model before training/testing/predicting. Use `model.build()`.')

        if isinstance(index, Sequence):
            test_data = index
        else:
            test_data = self.test_sequence(index)
            self.index_test = asintarr(index)

        if self.do_before_test is not None:
            self.do_before_test(**kwargs)

        loss, accuracy = self.do_forward(test_data, training=False)

        return loss, accuracy

    def do_forward(self, sequence, training=True):
        """
            Forward propagation for the input `sequence`. This method will be called 
            in `train` and `test`, you can rewrite it for you customized training/testing 
            process. If you want to specify your customized data during traing/testing/predicting, 
            you can implement a sub-class of `graphgallery.NodeSequence`, wich is iterable 
            and yields `inputs` and `labels` in each iteration. 


        Note:
        ----------
            You must compile your model before training/testing/predicting. 
            Use `model.build()`.

        Arguments:
        ----------
            sequence: `graphgallery.NodeSequence`
                The input `sequence`.    
            trainng (Boolean, optional): 
                Indicating training or test procedure. (default: :obj:`True`)

        Return:
        ----------
            loss: Float scalar
                Output loss of forward propagation.
            accuracy: Float scalar
                Output accuracy of prediction.

        """
        model = self.model

        if training:
            forward_fn = model.train_on_batch
        else:
            forward_fn = model.test_on_batch

        model.reset_metrics()

        with tf.device(self.device):
            for inputs, labels in sequence:
                loss, accuracy = forward_fn(x=inputs, y=labels, reset_metrics=False)

        return loss, accuracy

    def predict(self, index, **kwargs):
        """
            Predict the output probability for the `index` of nodes.

        Note:
        ----------
            You must compile your model before training/testing/predicting. 
            Use `model.build()`.

        Arguments:
        ----------
            index: Numpy array-like, `list` or integer scalar
                The index of nodes that will be computed.    

            **kwargs (optional): Additional arguments of
                :method:`do_before_predict`.   

        Return:
        ----------
            The predicted probability of each class for each node, 
            shape (len(index), n_classes).

        """

        if not self.model:
            raise RuntimeError('You must compile your model before training/testing/predicting. Use `model.build()`.')

        if self.do_before_predict is not None:
            self.do_before_predict(**kwargs)

    def train_sequence(self, index):
        """
            Construct the training sequence for the `index` of nodes.


        Arguments:
        ----------
            index: Numpy array-like, `list` or integer scalar
                The index of nodes used in training.

        Return:
        ----------
            The sequence of `graphgallery.NodeSequence` for the nodes.

        """

        raise NotImplementedError

    def test_sequence(self, index):
        """
            Construct the testing sequence for the `index` of nodes.

        Note:
        ----------
            If not implemented, this method will call `train_sequence` automatically.

        Arguments:
        ----------
            index: Numpy array-like, `list` or integer scalar
                The index of nodes used in testing.

        Return:
        ----------
            The sequence of `graphgallery.NodeSequence` for the nodes.
        """
        return self.train_sequence(index)

    def test_predict(self, index):
        """
            Predict the output accuracy for the `index` of nodes.

        Note:
        ----------
            You must compile your model before training/testing/predicting. 
            Use `model.build()`.

        Arguments:
        ----------
            index: Numpy array-like, `list` or integer scalar
                The index of nodes that will be computed.    

        Return:
        ----------
            accuracy: Float scalar
                The output accuracy of the `index` of nodes.

        """
        index = asintarr(index)
        logit = self.predict(index)
        predict_class = logit.argmax(1)
        labels = self.labels[index]
        return (predict_class == labels).mean()

#     @tf.function
    def __call__(self, inputs):
        return self.model(inputs)

    @property
    def weights(self):
        """Return the weights of model, type `tf.Tensor`."""
        return self.model.weights

    @property
    def np_weights(self):
        """Return the weights of model, type Numpy array-like."""
        return [weight.numpy() for weight in self.weights]

    @property
    def trainable_variables(self):
        """Return the trainable weights of model, type `tf.Tensor`."""
        return self.model.trainable_variables

    @property
    def close(self):
        """Close the session of model and set `built` to False."""
        self.set_model(None)
        K.clear_session()



import random
import numpy as np
import tensorflow as tf
import scipy.sparse as sp

from graphgallery.sequence.node_sequence import NodeSequence
from graphgallery.utils.misc import column_prop
from graphgallery.utils.graph_utils import sample_neighbors
from graphgallery import astensor


class ClusterMiniBatchSequence(NodeSequence):

    def __init__(
        self,
        inputs,
        labels,
        shuffle_batches=True,
        batch_size=1,
    ):
        assert batch_size == 1
        self.inputs, self.labels = astensor([inputs, labels])
        self.n_batches = len(self.inputs)
        self.shuffle_batches = shuffle_batches
        self.batch_size = batch_size
        self.indices = list(range(self.n_batches))

    def __len__(self):
        return self.n_batches

    def __getitem__(self, index):
        idx = self.indices[index]
        labels = self.labels[idx]
        return self.inputs[idx], labels

    def on_epoch_end(self):
        if self.shuffle_batches:
            self.shuffle()

    def shuffle(self):
        """
         Shuffle all nodes at the end of each epoch
        """
        random.shuffle(self.indices)


class SAGEMiniBatchSequence(NodeSequence):

    def __init__(
        self,
        inputs,
        labels=None,
        neighbors=None,
        n_samples=[5, 5],
        shuffle_batches=False,
        batch_size=512
    ):
        self.x, self.nodes = inputs
        self.labels = labels
        self.neighbors = neighbors
        self.n_batches = int(np.ceil(len(self.nodes)/batch_size))
        self.shuffle_batches = shuffle_batches
        self.batch_size = batch_size
        self.indices = np.arange(len(self.nodes))
        self.n_samples = n_samples

        self.x = astensor(self.x)

    def __len__(self):
        return self.n_batches

    def __getitem__(self, index):
        if self.shuffle_batches:
            idx = self.indices[index*self.batch_size:(index+1)*self.batch_size]
        else:
            idx = slice(index*self.batch_size, (index+1)*self.batch_size)

        nodes_input = [self.nodes[idx]]
        for n_sample in self.n_samples:
            neighbors = sample_neighbors(self.neighbors, nodes_input[-1], n_sample).ravel()
            nodes_input.append(neighbors)

        labels = self.labels[idx] if self.labels is not None else None

        return astensor([[self.x, *nodes_input], labels])

    def on_epoch_end(self):
        if self.shuffle_batches:
            self.shuffle()

    def shuffle(self):
        """
         Shuffle all nodes at the end of each epoch
        """
        random.shuffle(self.indices)


class FastGCNBatchSequence(NodeSequence):

    def __init__(
        self,
        inputs,
        labels=None,
        shuffle_batches=False,
        batch_size=None,
        rank=None
    ):
        self.x, self.adj = inputs
        self.labels = labels
        self.n_batches = int(np.ceil(self.adj.shape[0]/batch_size)) if batch_size else 1
        self.shuffle_batches = shuffle_batches
        self.batch_size = batch_size
        self.indices = np.arange(self.adj.shape[0])
        self.rank = rank
        if rank:
            self.p = column_prop(self.adj)

    def __len__(self):
        return self.n_batches

    def __getitem__(self, index):
        if self.batch_size is None:
            (x, adj), labels = self.full_batch(index)
        else:
            (x, adj), labels = self.mini_batch(index)

        if self.rank is not None:
            p = self.p
            rank = self.rank
            distr = adj.sum(0).A1.nonzero()[0]
            if rank > distr.size:
                q = distr
            else:
                q = np.random.choice(distr, rank, replace=False, p=p[distr]/p[distr].sum())
            adj = adj[:, q].dot(sp.diags(1.0 / (p[q] * rank)))

            if tf.is_tensor(x):
                x = tf.gather(x, q)
            else:
                x = x[q]

        return astensor([(x, adj), labels])

    def full_batch(self, index):
        return (self.x, self.adj), self.labels

    def mini_batch(self, index):
        if self.shuffle_batches:
            idx = self.indices[index*self.batch_size:(index+1)*self.batch_size]
        else:
            idx = slice(index*self.batch_size, (index+1)*self.batch_size)

        labels = self.labels[idx]
        adj = self.adj[idx]
        features = self.x

        return (features, adj), labels

    def on_epoch_end(self):
        if self.shuffle_batches:
            self.shuffle()

    def shuffle(self):
        """
         Shuffle all nodes at the end of each epoch
        """
        random.shuffle(self.indices)



import tensorflow as tf
from tensorflow.keras import Model, Input
from tensorflow.keras.layers import Dropout, Softmax
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import regularizers

from graphgallery.nn.layers import GraphConvFeature
from graphgallery.nn.models import SupervisedModel
from graphgallery.sequence import FullBatchNodeSequence
from graphgallery import astensor, asintarr, normalize_x, normalize_adj


class GCNF(SupervisedModel):
    """
        GCN + feature matrix

        Arguments:
        ----------
            adj: shape (N, N), Scipy sparse matrix if  `is_adj_sparse=True`, 
                Numpy array-like (or matrix) if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            x: shape (N, F), Scipy sparse matrix if `is_x_sparse=True`, 
                Numpy array-like (or matrix) if `is_x_sparse=False`.
                The input node feature matrix, where `F` is the dimension of features.
            labels: Numpy array-like with shape (N,)
                The ground-truth labels for all nodes in graph.
            norm_adj_rate (Float scalar, optional): 
                The normalize rate for adjacency matrix `adj`. (default: :obj:`-0.5`, 
                i.e., math:: \hat{A} = D^{-\frac{1}{2}} A D^{-\frac{1}{2}}) 
            norm_x_type (Boolean, optional): 
                Whether to use row-wise normalization for node feature matrix. 
                (default :bool: `True`)
            device (String, optional): 
                The device where the model is running on. You can specified `CPU` or `GPU` 
                for the model. (default: :str: `CPU:0`, i.e., running on the 0-th `CPU`)
            seed (Positive integer, optional): 
                Used in combination with `tf.random.set_seed` & `np.random.seed` & `random.seed`  
                to create a reproducible sequence of tensors across multiple calls. 
                (default :obj: `None`, i.e., using random seed)
            name (String, optional): 
                Specified name for the model. (default: :str: `class.__name__`)

    """

    def __init__(self, adj, x, labels, norm_adj_rate=-0.5, norm_x_type='l1',
                 device='CPU:0', seed=None, name=None, **kwargs):

        super().__init__(adj, x, labels, device=device, seed=seed, name=name, **kwargs)

        self.norm_adj_rate = norm_adj_rate
        self.norm_x_type = norm_x_type
        self.preprocess(adj, x)

    def preprocess(self, adj, x):
        super().preprocess(adj, x)
        # check the input adj and x, and convert them into proper data types
        adj, x = self._check_inputs(adj, x)

        if self.norm_adj_rate:
            adj = normalize_adj(adj, self.norm_adj_rate)

        if self.norm_x_type:
            x = normalize_x(x, norm=self.norm_x_type)

        with tf.device(self.device):
            self.x_norm, self.adj_norm = astensor([x, adj])

    def build(self, hiddens=[16], activations=['relu'], dropout=0.5,
              lr=0.01, l2_norm=5e-4, use_bias=False):

        assert len(hiddens) == len(activations), "The number of hidden units and " \
            "activation functions should be the same."

        with tf.device(self.device):

            x = Input(batch_shape=[None, self.n_features], dtype=self.floatx, name='features')
            adj = Input(batch_shape=[None, None], dtype=self.floatx, sparse=True, name='adj_matrix')
            index = Input(batch_shape=[None], dtype=self.intx, name='index')

            h = x
            for hid, activation in zip(hiddens, activations):
                h = GraphConvFeature(hid, use_bias=use_bias,
                                     activation=activation,
                                     concat=True,
                                     kernel_regularizer=regularizers.l2(l2_norm))([h, adj])

                h = Dropout(rate=dropout)(h)

            h = GraphConvFeature(self.n_classes, use_bias=use_bias)([h, adj])
            # To aviod the UserWarning of `tf.gather`, but it causes the shape
            # of the input data to remain the same
            # if ensure_shape:
            #     h = tf.ensure_shape(h, [self.n_nodes, self.n_classes])
            h = tf.gather(h, index)
            output = Softmax()(h)

            model = Model(inputs=[x, adj, index], outputs=output)
            model.compile(loss='sparse_categorical_crossentropy', optimizer=Adam(lr=lr), metrics=['accuracy'])

            self.set_model(model)

    def train_sequence(self, index):
        index = asintarr(index)
        labels = self.labels[index]
        with tf.device(self.device):
            sequence = FullBatchNodeSequence([self.x_norm, self.adj_norm, index], labels)
        return sequence

    def predict(self, index):
        super().predict(index)
        index = asintarr(index)
        with tf.device(self.device):
            index = astensor(index)
            logit = self.model.predict_on_batch([self.x_norm, self.adj_norm, index])

        if tf.is_tensor(logit):
            logit = logit.numpy()
        return logit



from numba import njit
import numpy as np

@njit
def mask_neighbors(indices, indptr, mask, nbrs, radius):
    mask[nbrs] = True
    if radius <= 1:
        return mask
    else:
        for n in nbrs:
            next_nbrs = indices[indptr[n]:indptr[n + 1]]
            mask = mask_neighbors(indices, indptr, mask, next_nbrs, radius-1)
        return mask
    
@njit
def find_4o_nbrs(indices, indptr, firstlevel, radius=4):
    N = len(indptr) - 1
    nodes = np.arange(N)
    for n in firstlevel:
        mask = np.asarray([False]*N)
        nbrs = indices[indptr[n]:indptr[n + 1]]
        mask = mask_neighbors(indices, indptr, mask, nbrs, radius)
        yield nodes[mask]



import numpy as np
import tensorflow as tf
import scipy.sparse as sp
from graphgallery.utils.data_utils import normalize_adj

# ChebyGCN
def chebyshev_polynomials(adj, rate=-0.5, order=3):
    """Calculate Chebyshev polynomials up to order k. Return a list of sparse matrices (tuple representation)."""

    assert order >= 2
    adj_normalized = normalize_adj(adj, rate=rate, add_self_loop=False)
    I = sp.eye(adj.shape[0], dtype=adj.dtype).tocsr()
    laplacian = I - adj_normalized
    largest_eigval = sp.linalg.eigsh(laplacian, 1, which='LM', return_eigenvectors=False)[0]
    scaled_laplacian = (2. / largest_eigval) * laplacian - I

    t_k = []
    t_k.append(I)
    t_k.append(scaled_laplacian)

    def chebyshev_recurrence(t_k_minus_one, t_k_minus_two, scaled_lap):
        return 2 * scaled_lap.dot(t_k_minus_one) - t_k_minus_two

    for i in range(2, order+1):
        t_k.append(chebyshev_recurrence(t_k[-1], t_k[-2], scaled_laplacian))

    return t_k

# FASTGCN
def column_prop(adj):
    column_norm = sp.linalg.norm(adj, axis=0)
    norm_sum = column_norm.sum()
    return column_norm/norm_sum

# LGCN
def solve_cudnn_error():
    gpus = tf.config.experimental.list_physical_devices('GPU')
    if gpus:
        try:
            # Currently, memory growth needs to be the same across GPUs
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, True)
            logical_gpus = tf.config.experimental.list_logical_devices('GPU')
            print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
        except RuntimeError as e:
            # Memory growth must be set before GPUs have been initialized
            print(e)


from graphgallery.utils.history import History
from graphgallery.utils.data_utils import Bunch, sample_mask, normalize_adj, normalize_x
from graphgallery.utils.shape_utils import repeat
from graphgallery.utils.tensor_utils import normalize_edge_tensor
from graphgallery.utils.gdc import GDC
from graphgallery.utils.probar import progress_bar


import numpy as np
import tensorflow as tf


from tensorflow.keras.utils import Sequence
from tensorflow.keras.layers import Layer


class NodeSequence(Sequence):

    def __init__(self, *args, **kwargs,):
        pass

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, index):
        raise NotImplementedError

    def on_epoch_end(self):
        pass

    def shuffle(self):
        pass




import tensorflow as tf
import numpy as np
import collections.abc as collections_abc


def is_list_like(x):
    """Check if x is list like: `List` or `Tuple`
    """
    return isinstance(x, collections_abc.Sequence)


def is_sparse(x):
    """Check whether `x` is sparse.

    Check whether an object is a `tf.sparse.SparseTensor` or
    `tf.compat.v1.SparseTensorValue`.

    Arguments:
        x: A python object to check.

    Returns:
        `True` iff `x` is a `tf.sparse.SparseTensor` or
        `tf.compat.v1.SparseTensorValue`.
    """
    return isinstance(x, (tf.sparse.SparseTensor, tf.sparse.SparseTensorValue))


def is_tensor_or_variable(x):
    """Check whether `x` is tf.Tensor or tf.Variable.

    Arguments:
        x: A python object to check.

    Returns:
        `True` iff `x` is a `tf.Tensor` or `tf.Variable`.
    """
    return tf.is_tensor(x) or isinstance(x, tf.Variable)


def is_interger_scalar(x):
    return isinstance(x, (int, np.int8,
                           np.int16,
                           np.int32,
                           np.int64,
                           np.uint8,
                           np.uint16,
                           np.uint32,
                           np.uint64,
                           ))



import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer


class Top_k_features(Layer):
    """
        Top_k_features layer as in:
        [Large-Scale Learnable Graph Convolutional Networks](https://arxiv.org/abs/1808.03965)
        Tensorflow 1.x implementation: https://github.com/divelab/lgcn

        `Top_k_features` implements the operation:
        Select the top-k features for each node and each feature dimension.
        And finally the selected features will concatenated with the input feature matrix along last dimension.

        Arguments:
          k: Positive Integer, Number of top elements to look for.

        Input shape:
          tuple/list with two 2-D tensor: Tensor `x` and SparseTensor `adj`: `[(n_nodes, n_features), (n_nodes, n_nodes)]`.
          The former one is the feature matrix (Tensor) and the other is adjacency matrix (SparseTensor).

        Output shape:
          3-D tensor with shape: `(n_nodes, k+1, n_features)`.
    """

    def __init__(self, k, **kwargs):

        super().__init__(**kwargs)
        self.k = k

    def call(self, inputs):

        x, adj = inputs
        if K.is_sparse(adj):
            adj = tf.sparse.to_dense(adj) # the adjacency matrix will be transformed into dense matrix
        adj = tf.expand_dims(adj, axis=1)  # (N, 1, N)
        x = tf.expand_dims(x, axis=-1)  # (N, F, 1)
        h = adj * x  # (N, F, N)
        h = tf.transpose(h, perm=(2, 1, 0))
        h = tf.math.top_k(h, k=self.k, sorted=True).values
        h = tf.concat([x, h], axis=-1)
        h = tf.transpose(h, perm=(0, 2, 1))
        return h  # (N, k+1, F)

    def get_config(self):
        config = {'k': self.k}

        base_config = super().get_config()
        return {**base_config, **config}

    def compute_output_shape(self, input_shapes):
        features_shape = input_shapes[0]
        output_shape = (features_shape[0], self.k+1, features_shape[1])
        return tf.TensorShape(output_shape)



import numpy as np
import tensorflow as tf
from tensorflow.keras import Model, Input
from tensorflow.keras.layers import Dropout, Softmax
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import regularizers

from graphgallery.nn.layers import MeanAggregator, GCNAggregator
from graphgallery.nn.models import SupervisedModel
from graphgallery.sequence import SAGEMiniBatchSequence
from graphgallery.utils.graph_utils import construct_neighbors
from graphgallery import astensor, asintarr, normalize_x


class GraphSAGE(SupervisedModel):
    """
        Implementation of SAmple and aggreGatE Graph Convolutional Networks (GraphSAGE). 
        `Inductive Representation Learning on Large Graphs <https://arxiv.org/abs/1706.02216>`
        Tensorflow 1.x implementation: <https://github.com/williamleif/GraphSAGE>
        Pytorch implementation: <https://github.com/williamleif/graphsage-simple/>


        Arguments:
        ----------
            adj: shape (N, N), Scipy sparse matrix if  `is_adj_sparse=True`, 
                Numpy array-like (or matrix) if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            x: shape (N, F), Scipy sparse matrix if `is_x_sparse=True`, 
                Numpy array-like (or matrix) if `is_x_sparse=False`.
                The input node feature matrix, where `F` is the dimension of features.
            labels: Numpy array-like with shape (N,)
                The ground-truth labels for all nodes in graph.
            n_samples (List of positive integer, optional): 
                The number of sampled neighbors for each nodes in each layer. 
                (default :obj: `[10, 5]`, i.e., sample `10` first-order neighbors and 
                `5` sencond-order neighbors, and the radius for `GraphSAGE` is `2`)
            norm_x_type (String, optional): 
                How to normalize the node feature matrix. See `graphgallery.normalize_x`
                (default :str: `l1`)
            device (String, optional): 
                The device where the model is running on. You can specified `CPU` or `GPU` 
                for the model. (default: :str: `CPU:0`, i.e., running on the 0-th `CPU`)
            seed (Positive integer, optional): 
                Used in combination with `tf.random.set_seed` & `np.random.seed` & `random.seed`  
                to create a reproducible sequence of tensors across multiple calls. 
                (default :obj: `None`, i.e., using random seed)
            name (String, optional): 
                Specified name for the model. (default: :str: `class.__name__`)


    """

    def __init__(self, adj, x, labels, n_samples=[15, 5], norm_x_type='l1',
                 device='CPU:0', seed=None, name=None, **kwargs):

        super().__init__(adj, x, labels, device=device, seed=seed, name=name, **kwargs)

        self.n_samples = n_samples
        self.norm_x_type = norm_x_type
        self.preprocess(adj, x)

    def preprocess(self, adj, x):
        super().preprocess(adj, x)
        # check the input adj and x, and convert them into proper data types
        adj, x = self._check_inputs(adj, x)

        if self.norm_x_type:
            x = normalize_x(x, norm=self.norm_x_type)

        # Dense matrix, shape [n_nodes, max_degree]
        neighbors = construct_neighbors(adj, max_degree=max(self.n_samples))
        # pad with a dummy zero vector
        x = np.vstack([x, np.zeros(self.n_features, dtype=self.floatx)])

        with tf.device(self.device):
            self.x_norm, self.neighbors = astensor(x), neighbors

    def build(self, hiddens=[64], activations=['relu'], dropout=0.5, lr=0.01, l2_norm=5e-4,
              output_normalize=False, aggrator='mean'):

        assert len(hiddens) == len(self.n_samples)-1, "The number of hidden units and " \
            "samples per layer should be the same"
        assert len(hiddens) == len(activations), "The number of hidden units and " \
            "activation functions should be the same."

        with tf.device(self.device):

            if aggrator == 'mean':
                Agg = MeanAggregator
            elif aggrator == 'gcn':
                Agg = GCNAggregator
            else:
                raise ValueError(f'Invalid value of `aggrator`, allowed values (`mean`, `gcn`), but got `{aggrator}`.')

            x = Input(batch_shape=[None, self.n_features], dtype=self.floatx, name='features')
            nodes = Input(batch_shape=[None], dtype=self.intx, name='nodes')
            neighbors = [Input(batch_shape=[None], dtype=self.intx, name=f'neighbors_{hop}')
                         for hop, n_sample in enumerate(self.n_samples)]

            aggrators = []
            for i, (hid, activation) in enumerate(zip(hiddens, activations)):
                # you can use `GCNAggregator` instead
                aggrators.append(Agg(hid, concat=True, activation=activation,
                                     kernel_regularizer=regularizers.l2(l2_norm)))

            aggrators.append(Agg(self.n_classes))

            h = [tf.nn.embedding_lookup(x, node) for node in [nodes, *neighbors]]
            for agg_i, aggrator in enumerate(aggrators):
                feature_shape = h[0].shape[-1]
                for hop in range(len(self.n_samples)-agg_i):
                    neighbor_shape = [-1, self.n_samples[hop], feature_shape]
                    h[hop] = Dropout(rate=dropout)(aggrator([h[hop], tf.reshape(h[hop+1], neighbor_shape)]))
                h.pop()

            h = h[0]
            if output_normalize:
                h = tf.nn.l2_normalize(h, axis=1)
            output = Softmax()(h)

            model = Model(inputs=[x, nodes, *neighbors], outputs=output)
            model.compile(loss='sparse_categorical_crossentropy', optimizer=Adam(lr=lr), metrics=['accuracy'])

            self.set_model(model)

    def train_sequence(self, index):
        index = asintarr(index)
        labels = self.labels[index]
        with tf.device(self.device):
            sequence = SAGEMiniBatchSequence([self.x_norm, index], labels, neighbors=self.neighbors, n_samples=self.n_samples)
        return sequence

    def predict(self, index):
        super().predict(index)
        logit = []
        index = asintarr(index)
        with tf.device(self.device):
            data = SAGEMiniBatchSequence([self.x_norm, index], neighbors=self.neighbors, n_samples=self.n_samples)
            for inputs, labels in data:
                output = self.model.predict_on_batch(inputs)
                if tf.is_tensor(output):
                    output = output.numpy()

                logit.append(output)
        logit = np.concatenate(logit, axis=0)
        return logit



import numpy as np
import scipy.sparse as sp
from sklearn.preprocessing import normalize
from scipy.linalg import expm

from graphgallery.utils.data_utils import normalize_adj

def clip_matrix(matrix, threshold: float) -> sp.csr_matrix:
    '''Sparsify using threshold epsilon'''
    assert sp.isspmatrix(matrix), 'Input matrix should be sparse matrix with format scipy.sparse.*_matrix.'
    matrix = matrix.tocsr()
    thres = np.vectorize(lambda x : x if x >= threshold else 0.)
    matrix.data = thres(matrix.data)
    matrix.eliminate_zeros()
    return matrix

def top_k_matrix(matrix, k: int) -> sp.csr_matrix:
    '''Row-wise select top-k values'''
    assert sp.isspmatrix(matrix), 'Input matrix should be sparse matrix with format scipy.sparse.*_matrix.'
    matrix = matrix.tolil()
    data = matrix.data
    for row in range(matrix.shape[0]):
        t = np.asarray(data[row])
        t[np.argsort(-t)[k:]] = 0.
        data[row] = t.tolist()
    matrix = matrix.tocsr()
    matrix.eliminate_zeros()
    return matrix

def GDC(adj: sp.csr_matrix, alpha: float = None, t: float = None, eps: float = None, k: int = None, which: str = 'PPR') -> sp.csr_matrix:
    
    if not (eps or k):
        raise RuntimeError('Either `eps` or `k` should be specified!')
    if eps and k:
        raise RuntimeError('Only one of `eps` and `k` should be specified!')

    N = adj.shape[0]
    
    # Symmetric transition matrix
    T_sym = normalize_adj(adj)
    
    if which == 'PPR':
        # PPR-based diffusion
        assert alpha is not None, '`alpha` should be specified for PPR-based diffusion.'
        S = alpha * sp.linalg.inv((sp.eye(N) - (1 - alpha) * T_sym).tocsc())
    elif which == 'Heat':
        assert t is not None, '`t` should be specified for Heat-based diffusion.'
        S = -t * (sp.eye(N) - T_sym)
        S = expm(S.toarray())
        S = sp.csr_matrix(S)
    else:
        raise ValueError(f'Invalid argument of `{which}`.')
        
    if eps:
        # Sparsify using threshold epsilon
        S = clip_matrix(S, threshold=eps)
    else:
        # Row-wise select top-k values
        S = top_k_matrix(S, k=k)

    # Column-normalized transition matrix on graph S_tilde
    T_S = normalize(S, norm='l1', axis=0)

    return T_S.tocsr()



try:
    import texttable
except ImportError:
    texttable = None
import numpy as np
from collections import defaultdict


class History:
    
    def __init__(self, monitor_metric='val_acc', early_stop_metric='val_loss'):
        
        self._check_name(monitor_metric)
        self._check_name(early_stop_metric)
        self.monitor_metric = monitor_metric
        self.early_stop_metric = early_stop_metric
                
        self._history = None
        self.epoch = None
        self.times = None
        self.patience = None
        self.best_results = None
        self.save_best = None
        self.early_stop_at = None
        
        self.reset()
        
    def reset(self):
        self._history = defaultdict(list)
        self.epoch = []
        self.times = 0
        self.patience = {'loss':0, 'val_loss':0, 
                         'acc': 0, 'val_acc': 0}
        self.best_results = {'loss': np.inf, 'val_loss': np.inf, 
                             'acc': -np.inf, 'val_acc': -np.inf}
        self.save_best = False
        self.early_stop_at = None
        
    def register_monitor_metric(self, name):
        self._check_name(name)
        self.monitor_metric = name
        
    def register_early_stop_metric(self, name):
        self._check_name(name)
        self.early_stop_metric = name      
        
    def add_results(self, value, name):
        self._check_name(name)
        self.save_best_func(value, name)
        self.early_stop_func(value, name)
            
        self._history[name].append(value)
        
    def save_best_func(self, value, name):
        
        self.save_best = False
        
        if (name.endswith('acc') and value > self.best_results[name] or
            name.endswith('loss') and value < self.best_results[name]):
            self.best_results[name] = value

            if name == self.monitor_metric:
                self.save_best = True
    
    def early_stop_func(self, value, name):
        
        if self.times == 0:
            return 
        
        if (name.endswith('acc') and value < self._history[name][-1] or
            name.endswith('loss') and value > self._history[name][-1]):
            self.patience[name] -= 1
        else:
            self.patience[name] = 0
        
#     def latest_results(self, n, name='val_loss', method='mean'):
#         Method = {'mean': np.mean, 'max': np.max, 'min': np.min, 'sum': np.sum}[method]
#         return Method(self._history[name][-(n+1):-1])
    
    def record_epoch(self, epoch):
        self.epoch.append(epoch)
        self.times += 1
        
    def time_to_early_stopping(self, early_stopping_patience):
        
        name = self.early_stop_metric
        if self.patience[name] + early_stopping_patience <=0:
            self.early_stop_at = self.epoch[-1]
            return True
        else:
            return False
        
    @staticmethod    
    def _check_name(name):
        assert name in ('loss', 'val_loss', 'acc', 'val_acc')
        
    @property
    def history(self):
        return dict(self._history)
        
    def show(self):
        assert texttable is not None, "Please install `texttable` package!"
        t = texttable.Texttable()
        t.add_rows([["Training Details", "Value"],
                    ['Running times', self.times],
                    ['Lowerst training loss', self.best_results['loss']],
                    ['Highest training accuracy', self.best_results['acc']],
                    ['Lowerst validation loss', self.best_results['val_loss']],
                    ['Highest validation accuracy', self.best_results['val_acc']],
                    ['Metric to save best weights', self.monitor_metric],
                    ['Metric to early stopping', self.early_stop_metric],
                    ['Early stopping at', f'Epoch: {self.early_stop_at}']])
        
        print(t.draw())  



import numpy as np
import tensorflow as tf

from graphgallery import astensor, sample_mask
from graphgallery.sequence.node_sequence import NodeSequence


class NodeSampleSequence(NodeSequence):

    def __init__(
        self,
        inputs,
        labels,
        neighbors,
        n_samples=100,
        resample=True
    ):

        self.inputs = inputs
        self.labels = labels
        self.n_batches = 1
        self.neighbors = neighbors
        self.n_nodes = inputs[0].shape[0]
        self.n_samples = n_samples
        self.adv_mask = self.smple_nodes()
        self.resample = resample

    def __len__(self):
        return self.n_batches

    def __getitem__(self, index):

        item = [*self.inputs, self.adv_mask], self.labels
        if self.resample:
            self.adv_mask = self.smple_nodes()
        return astensor(item)

    def smple_nodes(self):
        N = self.n_nodes
        flag = np.zeros(N, dtype=np.bool)
        adv_index = np.zeros(self.n_samples, dtype='int32')
        for i in range(self.n_samples):
            n = np.random.randint(0, N)
            while flag[n]:
                n = np.random.randint(0, N)
            adv_index[i] = n
            flag[self.neighbors[n]] = True
            if flag.sum() == N:
                break
        adv_mask = np.zeros(N, dtype='float32')
        adv_mask[adv_index] = 1.
        return adv_mask



from graphgallery.sequence.node_sequence import NodeSequence
from graphgallery.sequence.minibatch_sequence import ClusterMiniBatchSequence, SAGEMiniBatchSequence, FastGCNBatchSequence
from graphgallery.sequence.fullbatch_sequence import FullBatchNodeSequence
from graphgallery.sequence.sample_sequence import NodeSampleSequence


from tensorflow.keras import activations, constraints, initializers, regularizers
from tensorflow.keras.layers import Layer, Dropout, LeakyReLU

import tensorflow as tf


class DenseGraphConv(Layer):
    """
        Basic graph convolution layer as in: 
        `Semi-Supervised Classification with Graph Convolutional Networks` (https://arxiv.org/abs/1609.02907)
        Tensorflow 1.x implementation: https://github.com/tkipf/gcn
        Pytorch implementation: https://github.com/tkipf/pygcn

        `DenseGraphConv` implements the `Dense` operation:
        `output = activation(adj @ x @ kernel + bias)`
        where `x` is the feature matrix, `adj` is the adjacency matrix with dense form,
        `activation` is the element-wise activation function
        passed as the `activation` argument, `kernel` is a weights matrix
        created by the layer, and `bias` is a bias vector created by the layer
        (only applicable if `use_bias` is `True`).


        Arguments:
          units: Positive integer, dimensionality of the output space.
          activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
          use_bias: Boolean, whether the layer uses a bias vector.
          kernel_initializer: Initializer for the `kernel` weights matrix.
          bias_initializer: Initializer for the bias vector.
          kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
          bias_regularizer: Regularizer function applied to the bias vector.
          activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
          kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
          bias_constraint: Constraint function applied to the bias vector.

        Input shape:
          tuple/list with two 2-D tensor: Tensor `x` and SparseTensor `adj`: `[(n_nodes, n_features), (n_nodes, n_nodes)]`.
          The former one is the feature matrix (Tensor) and the last is adjacency matrix (Tensor).

        Output shape:
          2-D tensor with shape: `(n_nodes, units)`.       
    """

    def __init__(self, units,
                 use_bias=False,
                 activation=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):

        super().__init__(**kwargs)
        self.units = units
        self.use_bias = use_bias

        self.activation = activations.get(activation)
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

    def build(self, input_shapes):
        self.kernel = self.add_weight(shape=(input_shapes[0][-1], self.units),
                                      initializer=self.kernel_initializer,
                                      name='kernel',
                                      regularizer=self.kernel_regularizer,
                                      constraint=self.kernel_constraint)
        if self.use_bias:
            self.bias = self.add_weight(shape=(self.units,),
                                        initializer=self.bias_initializer,
                                        name='bias',
                                        regularizer=self.bias_regularizer,
                                        constraint=self.bias_constraint)
        else:
            self.bias = None

        super().build(input_shapes)

    def call(self, inputs):

        x, adj = inputs
        h = x @ self.kernel
        output = adj @ h

        if self.use_bias:
            output += self.bias

        return self.activation(output)

    def get_config(self):
        config = {'units': self.units,
                  'use_bias': self.use_bias,
                  'activation': activations.serialize(self.activation),
                  'kernel_initializer': initializers.serialize(
                      self.kernel_initializer),
                  'bias_initializer': initializers.serialize(
                      self.bias_initializer),
                  'kernel_regularizer': regularizers.serialize(
                      self.kernel_regularizer),
                  'bias_regularizer': regularizers.serialize(
                      self.bias_regularizer),
                  'activity_regularizer': regularizers.serialize(
                      self.activity_regularizer),
                  'kernel_constraint': constraints.serialize(
                      self.kernel_constraint),
                  'bias_constraint': constraints.serialize(self.bias_constraint)
                  }

        base_config = super().get_config()
        return {**base_config, **config}

    def compute_output_shape(self, input_shapes):
        features_shape = input_shapes[0]
        output_shape = (features_shape[0], self.units)
        return tf.TensorShape(output_shape)  # (n_nodes, output_dim)



import tensorflow as tf
from tensorflow.keras import Model, Input
from tensorflow.keras.layers import Dropout, Softmax
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import regularizers

from graphgallery.nn.layers import WaveletConvolution
from graphgallery.nn.models import SupervisedModel
from graphgallery.sequence import FullBatchNodeSequence
from graphgallery.utils.wavelet_utils import wavelet_basis
from graphgallery import astensor, asintarr, normalize_x


class GWNN(SupervisedModel):
    """
        Implementation of Graph Wavelet Neural Networks (GWNN). 
        `Graph Wavelet Neural Network <https://arxiv.org/abs/1904.07785>`
        Tensorflow 1.x implementation: <https://github.com/Eilene/GWNN>
        Pytorch implementation: <https://github.com/benedekrozemberczki/GraphWaveletNeuralNetwork>

        Arguments:
        ----------
            adj: shape (N, N), Scipy sparse matrix if  `is_adj_sparse=True`, 
                Numpy array-like (or matrix) if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            x: shape (N, F), Scipy sparse matrix if `is_x_sparse=True`, 
                Numpy array-like (or matrix) if `is_x_sparse=False`.
                The input node feature matrix, where `F` is the dimension of features.
            labels: Numpy array-like with shape (N,)
                The ground-truth labels for all nodes in graph.
            order (Positive integer, optional): 
                The power (order) of approximated wavelet matrix using Chebyshev polynomial 
                filter. (default :obj: `3`)
            wavelet_s (Float scalar, optional): 
                The wavelet constant corresponding to a heat kernel. 
                (default: :obj:`1.2`) 
            threshold (Float scalar, optional): 
                Used to sparsify the wavelet matrix. (default: :obj:`1e-4`, i.e., 
                values less than `1e-4` will be set to zero to preserve the sparsity 
                of wavelet matrix)       
            wavelet_normalize (Boolean, optional): 
                Whether to use row-normalize for wavelet matrix. (default :bool: `True`)
            norm_x_type (String, optional): 
                How to normalize the node feature matrix. See `graphgallery.normalize_x`
                (default :str: `l1`)
            device (String, optional): 
                The device where the model is running on. You can specified `CPU` or `GPU` 
                for the model. (default: :str: `CPU:0`, i.e., running on the 0-th `CPU`)
            seed (Positive integer, optional): 
                Used in combination with `tf.random.set_seed` & `np.random.seed` & `random.seed`  
                to create a reproducible sequence of tensors across multiple calls. 
                (default :obj: `None`, i.e., using random seed)
            name (String, optional): 
                Specified name for the model. (default: :str: `class.__name__`)
    """

    def __init__(self, adj, x, labels, order=3, wavelet_s=1.2,
                 threshold=1e-4, wavelet_normalize=True, norm_x_type='l1',
                 device='CPU:0', seed=None, name=None, **kwargs):

        super().__init__(adj, x, labels, device=device, seed=seed, name=name, **kwargs)

        self.order = order
        self.wavelet_s = wavelet_s
        self.threshold = threshold
        self.wavelet_normalize = wavelet_normalize
        self.norm_x_type = norm_x_type
        self.preprocess(adj, x)

    def preprocess(self, adj, x):
        super().preprocess(adj, x)
        # check the input adj and x, and convert them into proper data types
        adj, x = self._check_inputs(adj, x)

        if self.norm_x_type:
            x = normalize_x(x, norm=self.norm_x_type)

        wavelet, inverse_wavelet = wavelet_basis(adj, wavelet_s=self.wavelet_s,
                                                 order=self.order, threshold=self.threshold,
                                                 wavelet_normalize=self.wavelet_normalize)
        with tf.device(self.device):
            self.x_norm, self.adj_norm = astensor([x, [wavelet, inverse_wavelet]])

    def build(self, hiddens=[16], activations=['relu'], dropout=0.5, lr=0.01, l2_norm=5e-4):

        assert len(hiddens) == len(activations), "The number of hidden units and " \
            "activation functions should be the same."

        with tf.device(self.device):

            x = Input(batch_shape=[self.n_nodes, self.n_features], dtype=self.floatx, name='features')
            wavelet = Input(batch_shape=[self.n_nodes, self.n_nodes], dtype=self.floatx, sparse=True, name='wavelet')
            inverse_wavelet = Input(batch_shape=[self.n_nodes, self.n_nodes], dtype=self.floatx, sparse=True,
                                    name='inverse_wavelet')
            index = Input(batch_shape=[None],  dtype=self.intx, name='index')

            h = x

            for hid, activation in zip(hiddens, activations):
                h = WaveletConvolution(hid, activation=activation,
                                       kernel_regularizer=regularizers.l2(l2_norm))([h, wavelet, inverse_wavelet])
                h = Dropout(rate=dropout)(h)

            h = WaveletConvolution(self.n_classes)([h, wavelet, inverse_wavelet])
            h = tf.ensure_shape(h, [self.n_nodes, self.n_classes])
            h = tf.gather(h, index)
            output = Softmax()(h)

            model = Model(inputs=[x, wavelet, inverse_wavelet, index], outputs=output)
            model.compile(loss='sparse_categorical_crossentropy', optimizer=Adam(lr=lr), metrics=['accuracy'])

            self.set_model(model)

    def train_sequence(self, index):
        index = asintarr(index)
        labels = self.labels[index]

        with tf.device(self.device):
            sequence = FullBatchNodeSequence([self.x_norm, *self.adj_norm, index], labels)
        return sequence

    def predict(self, index):
        super().predict(index)
        index = asintarr(index)
        with tf.device(self.device):
            index = astensor(index)
            logit = self.model.predict_on_batch([self.x_norm, *self.adj_norm, index])

        if tf.is_tensor(logit):
            logit = logit.numpy()
        return logit



import numpy as np
import networkx as nx

from gensim.models import Word2Vec

from graphgallery.nn.models import UnsupervisedModel
from graphgallery.utils.walker import RandomWalker, alias_sample


class Node2vec(UnsupervisedModel):
    """
        Implementation of Node2vec Unsupervised Graph Neural Networks (Node2vec). 
        [node2vec: Scalable Feature Learning for Networks](https://arxiv.org/abs/1607.00653)
        Implementation: https://github.com/aditya-grover/node2vec
        Cpp implementation: https://github.com/snap-stanford/snap/tree/master/examples/node2vec

        Arguments:
        ----------
            adj: shape (N, N), Scipy sparse matrix if  `is_adj_sparse=True`, 
                Numpy array-like (or matrix) if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            x: shape (N, F), Scipy sparse matrix if `is_x_sparse=True`, 
                Numpy array-like (or matrix) if `is_x_sparse=False`.
                The input node feature matrix, where `F` is the dimension of features.
            labels: Numpy array-like with shape (N,)
                The ground-truth labels for all nodes in graph.
            device (String, optional): 
                The device where the model is running on. You can specified `CPU` or `GPU` 
                for the model. (default: :str: `CPU:0`, i.e., running on the 0-th `CPU`)
            seed (Positive integer, optional): 
                Used in combination with `tf.random.set_seed` & `np.random.seed` & `random.seed`  
                to create a reproducible sequence of tensors across multiple calls. 
                (default :obj: `None`, i.e., using random seed)
            name (String, optional): 
                Specified name for the model. (default: :str: `class.__name__`)

    """

    def __init__(self, adj, x, labels, graph=None, device='CPU:0', seed=None, name=None, **kwargs):

        super().__init__(adj, x, labels, device=device, seed=seed, name=name, **kwargs)

        if graph is None:
            graph = nx.from_scipy_sparse_matrix(adj, create_using=nx.DiGraph)

        self.walker = None
        self.graph = graph

    def build(self, walk_length=80, walks_per_node=10,
              embedding_dim=64, window_size=5, workers=16,
              iter=1, num_neg_samples=1, p=0.5, q=0.5):

        self.walker = RandomWalker(self.graph, p=p, q=q)
        self.walker.preprocess_transition_probs()

        walks = self.node2vec_random_walk(self.graph,
                                          self.walker.alias_nodes,
                                          self.walker.alias_edges,
                                          walk_length=walk_length,
                                          walks_per_node=walks_per_node)

        sentences = [list(map(str, walk)) for walk in walks]

        model = Word2Vec(sentences, size=embedding_dim, window=window_size, min_count=0, sg=1, workers=workers,
                         iter=iter, negative=num_neg_samples, hs=0, compute_loss=True)

        self.set_model(model)

    @staticmethod
    def node2vec_random_walk(G, alias_nodes, alias_edges, walk_length=80, walks_per_node=10):
        for _ in range(walks_per_node):
            for n in G.nodes():
                single_walk = [n]
                current_node = n
                for _ in range(walk_length-1):
                    neighbors = list(G.neighbors(current_node))
                    if len(neighbors) > 0:
                        if len(single_walk) == 1:
                            current_node = neighbors[alias_sample(alias_nodes[current_node][0], alias_nodes[current_node][1])]
                        else:
                            prev = single_walk[-2]
                            edge = (prev, current_node)
                            current_node = neighbors[alias_sample(alias_edges[edge][0], alias_edges[edge][1])]
                    else:
                        break

                    single_walk.append(current_node)
                yield single_walk

    def get_embeddings(self, norm=True):
        embeddings = self.model.wv.vectors[np.fromiter(map(int, self.model.wv.index2word), np.int32).argsort()]

        if norm:
            embeddings = self.normalize_embedding(embeddings)

        self.embeddings = embeddings



from tensorflow.keras import activations, constraints, initializers, regularizers
from tensorflow.keras.layers import Layer

import tensorflow as tf


class GraphConvFeature(Layer):

    def __init__(self, units,
                 use_bias=False,
                 activation=None,
                 concat=False,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):

        super().__init__(**kwargs)
        self.units = units
        self.use_bias = use_bias
        self.concat = concat
        self.output_dim = units*2 if concat else units

        self.activation = activations.get(activation)
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

    def build(self, input_shapes):
        self.kernel = self.add_weight(shape=(input_shapes[0][-1], self.units),
                                      initializer=self.kernel_initializer,
                                      name='kernel',
                                      regularizer=self.kernel_regularizer,
                                      constraint=self.kernel_constraint)
        if self.use_bias:
            self.bias = self.add_weight(shape=(self.output_dim,),
                                        initializer=self.bias_initializer,
                                        name='bias',
                                        regularizer=self.bias_regularizer,
                                        constraint=self.bias_constraint)
        else:
            self.bias = None

        super().build(input_shapes)

    def call(self, inputs):

        x, adj = inputs
        h = x @ self.kernel
        output = tf.sparse.sparse_dense_matmul(adj, h)

        if self.concat:
            output = tf.concat([output, h], axis=1)
        else:
            output = tf.divide(tf.add(output, h), 2.0)

        if self.use_bias:
            output += self.bias

        return self.activation(output)

    def get_config(self):
        config = {'units': self.units,
                  'use_bias': self.use_bias,
                  'concat': self.concat,
                  'activation': activations.serialize(self.activation),
                  'kernel_initializer': initializers.serialize(
                      self.kernel_initializer),
                  'bias_initializer': initializers.serialize(
                      self.bias_initializer),
                  'kernel_regularizer': regularizers.serialize(
                      self.kernel_regularizer),
                  'bias_regularizer': regularizers.serialize(
                      self.bias_regularizer),
                  'activity_regularizer': regularizers.serialize(
                      self.activity_regularizer),
                  'kernel_constraint': constraints.serialize(
                      self.kernel_constraint),
                  'bias_constraint': constraints.serialize(self.bias_constraint)
                  }

        base_config = super().get_config()
        return {**base_config, **config}

    def compute_output_shape(self, input_shapes):
        features_shape = input_shapes[0]
        output_shape = (features_shape[0], self.output_dim)
        return tf.TensorShape(output_shape)  # (n_nodes, output_dim)



"""Inspired by Keras backend config API. https://tensorflow.google.com """


from tensorflow.keras import backend as K



floatx = K.floatx
set_floatx = K.set_floatx
epsilon = K.epsilon
set_epsilon = K.set_epsilon

_INTX = 'int64'

def intx():
    """Returns the default integer type, as a string.

    E.g. `'int16'`, `'int32'`, `'int64'`.

    Returns:
      String, the current default integer type.

    Example:
    >>> graphgallery.config.intx()
    'int64'
    """       
    return _INTX


def set_intx(dtype):
    """Sets the default integer type.

    Arguments:
      value: String; `'int16'`, `'int32'`, or `'int64'`.

    Example:
    >>> graphgallery.config.intx()
    'int64'
    >>> graphgallery.config.set_intx('int32')
    >>> graphgallery.config.intx()
    'int32'

    Raises:
      ValueError: In case of invalid value.
    """ 
    
    if dtype not in {'int16', 'int32', 'int64'}:
        raise ValueError('Unknown floatx type: ' + str(dtype))
    global _INTX
    _INTX = dtype




import tensorflow as tf
from tensorflow.keras import Model, Input
from tensorflow.keras.layers import Dropout, Softmax
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import regularizers

from graphgallery.nn.layers import GaussionConvolution_F, GaussionConvolution_D
from graphgallery.nn.models import SupervisedModel
from graphgallery.sequence import FullBatchNodeSequence
from graphgallery import astensor, asintarr, normalize_x, normalize_adj


class RobustGCN(SupervisedModel):
    """
        Implementation of Robust Graph Convolutional Networks (RobustGCN). 
        `Robust Graph Convolutional Networks Against Adversarial Attacks 
        <https://dl.acm.org/doi/10.1145/3292500.3330851>`
        Tensorflow 1.x implementation: <https://github.com/thumanlab/nrlweb/blob/master/static/assets/download/RGCN.zip>

        Arguments:
        ----------
            adj: shape (N, N), Scipy sparse matrix if  `is_adj_sparse=True`, 
                Numpy array-like (or matrix) if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            x: shape (N, F), Scipy sparse matrix if `is_x_sparse=True`, 
                Numpy array-like (or matrix) if `is_x_sparse=False`.
                The input node feature matrix, where `F` is the dimension of features.
            labels: Numpy array-like with shape (N,)
                The ground-truth labels for all nodes in graph.
            norm_adj_rate (List of float scalar, optional): 
                The normalize rate for adjacency matrix `adj`. 
                (default: :obj:`[-0.5, -1]`, i.e., two normalized `adj` with rate `-0.5` 
                and `-1.0`, respectively) 
            norm_x_type (String, optional): 
                How to normalize the node feature matrix. See `graphgallery.normalize_x`
                (default :str: `l1`)
            device (String, optional): 
                The device where the model is running on. You can specified `CPU` or `GPU` 
                for the model. (default: :str: `CPU:0`, i.e., running on the 0-th `CPU`)
            seed (Positive integer, optional): 
                Used in combination with `tf.random.set_seed` & `np.random.seed` & `random.seed`  
                to create a reproducible sequence of tensors across multiple calls. 
                (default :obj: `None`, i.e., using random seed)
            name (String, optional): 
                Specified name for the model. (default: :str: `class.__name__`)

    """

    def __init__(self, adj, x, labels, norm_adj_rate=[-0.5, -1], norm_x_type='l1',
                 device='CPU:0', seed=None, name=None, **kwargs):

        super().__init__(adj, x, labels, device=device, seed=seed, name=name, **kwargs)

        self.norm_adj_rate = norm_adj_rate
        self.norm_x_type = norm_x_type
        self.preprocess(adj, x)

    def preprocess(self, adj, x):
        super().preprocess(adj, x)
        # check the input adj and x, and convert them into proper data types
        adj, x = self._check_inputs(adj, x)

        if self.norm_adj_rate:
            adj = normalize_adj([adj, adj], self.norm_adj_rate)    # [adj_1, adj_2]

        if self.norm_x_type:
            x = normalize_x(x, norm=self.norm_x_type)

        with tf.device(self.device):
            self.x_norm, self.adj_norm = astensor([x, adj])

    def build(self, hiddens=[64], activations=['relu'], use_bias=False, dropout=0.6, lr=0.01, l2_norm=1e-4, para_kl=5e-4, gamma=1.0, ensure_shape=True):

        assert len(hiddens) == len(activations), "The number of hidden units and " \
            "activation functions should be the same."

        with tf.device(self.device):

            x = Input(batch_shape=[None, self.n_features], dtype=self.floatx, name='features')
            adj = [Input(batch_shape=[None, None], dtype=self.floatx, sparse=True, name='adj_matrix_1'),
                   Input(batch_shape=[None, None], dtype=self.floatx, sparse=True, name='adj_matrix_2')]
            index = Input(batch_shape=[None],  dtype=self.intx, name='index')

            h = Dropout(rate=dropout)(x)
            h, KL_divergence = GaussionConvolution_F(hiddens[0], gamma=gamma,
                                                     use_bias=use_bias,
                                                     activation=activations[0],
                                                     kernel_regularizer=regularizers.l2(l2_norm))([h, *adj])

            # additional layers (usually unnecessay)
            for hid, activation in zip(hiddens[1:], activations[1:]):
                h = Dropout(rate=dropout)(h)
                h = GaussionConvolution_D(hid, gamma=gamma, use_bias=use_bias, activation=activation)([h, *adj])

            h = Dropout(rate=dropout)(h)
            h = GaussionConvolution_D(self.n_classes, gamma=gamma, use_bias=use_bias)([h, *adj])
            # To aviod the UserWarning of `tf.gather`, but it causes the shape
            # of the input data to remain the same
            if ensure_shape:
                h = tf.ensure_shape(h, [self.n_nodes, self.n_classes])
            h = tf.gather(h, index)
            output = Softmax()(h)

            model = Model(inputs=[x, *adj, index], outputs=output)
            model.compile(loss='sparse_categorical_crossentropy', optimizer=Adam(lr=lr), metrics=['accuracy'])
            model.add_loss(para_kl * KL_divergence)

            self.set_model(model)

    def train_sequence(self, index):
        index = asintarr(index)
        labels = self.labels[index]
        with tf.device(self.device):
            sequence = FullBatchNodeSequence([self.x_norm, *self.adj_norm, index], labels)
        return sequence

    def predict(self, index):
        super().predict(index)
        index = asintarr(index)

        with tf.device(self.device):
            index = astensor(index)
            logit = self.model.predict_on_batch([self.x_norm, *self.adj_norm, index])

        if tf.is_tensor(logit):
            logit = logit.numpy()
        return logit



from tensorflow.keras import activations, constraints, initializers, regularizers
from tensorflow.keras.layers import Layer, Dropout, LeakyReLU

import tensorflow as tf


class GraphAttention(Layer):
    """
        Basic graph attention layer as in: 
        [Graph Attention Networks](https://arxiv.org/abs/1710.10903)
        Tensorflow 1.x implementation: https://github.com/PetarV-/GAT
        Pytorch implementation: https://github.com/Diego999/pyGAT
        Keras implementation: https://github.com/danielegrattarola/keras-gat


        Arguments:
          units: Positive integer, dimensionality of the output space.
          atten_heads: Positive integer, number of attention heads.
          attn_heads_reduction: {'concat', 'average'}, whether to enforce concat or average for the outputs from different heads.
          dropout: Float, internal dropout rate
          activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
          use_bias: Boolean, whether the layer uses a bias vector.
          kernel_initializer: Initializer for the `kernel` weights matrix.
          attn_kernel_initializer: Initializer for the `attn_kernel` weights matrix.
          bias_initializer: Initializer for the bias vector.
          kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
          attn_kernel_regularizer: Regularizer function applied to
            the `attn_kernel` weights matrix.
          bias_regularizer: Regularizer function applied to the bias vector.
          activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation").
          kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
          attn_kernel_constraint: Constraint function applied to
            the `attn_kernel` weights matrix.
          bias_constraint: Constraint function applied to the bias vector.

        Input shape:
          tuple/list with two 2-D tensor: Tensor `x` and SparseTensor `adj`: `[(n_nodes, n_features), (n_nodes, n_nodes)]`.
          The former one is the feature matrix (Tensor) and the last is adjacency matrix (SparseTensor).

        Output shape:
          2-D tensor with shape: `(n_nodes, units)` (use average) or `(n_nodes, attn_heads * units)` (use concat).       
    """

    def __init__(self,
                 units,
                 attn_heads=1,
                 attn_heads_reduction='concat',
                 dropout=0.5,
                 activation=None,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 attn_kernel_initializer='glorot_uniform',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 attn_kernel_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 attn_kernel_constraint=None,
                 **kwargs):

        super().__init__(**kwargs)

        if attn_heads_reduction not in {'concat', 'average'}:
            raise ValueError('Possbile reduction methods: concat, average')

        self.units = units  # Number of output features (F' in the paper)
        self.attn_heads = attn_heads  # Number of attention heads (K in the paper)
        self.attn_heads_reduction = attn_heads_reduction  # Eq. 5 and 6 in the paper
        self.dropout = dropout  # Internal dropout rate
        self.activation = activations.get(activation)  # Eq. 4 in the paper
        self.use_bias = use_bias

        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.attn_kernel_initializer = initializers.get(attn_kernel_initializer)

        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.attn_kernel_regularizer = regularizers.get(attn_kernel_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)

        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        self.attn_kernel_constraint = constraints.get(attn_kernel_constraint)

        # Populated by build()
        self.kernels = []       # Layer kernels for attention heads
        self.biases = []        # Layer biases for attention heads
        self.attn_kernels = []  # Attention kernels for attention heads

        if attn_heads_reduction == 'concat':
            # Output will have shape (..., K * F')
            self.output_dim = self.units * self.attn_heads
        else:
            # Output will have shape (..., F')
            self.output_dim = self.units

    def build(self, input_shape):
        input_dim = input_shape[0][-1]

        # Initialize weights for each attention head
        for head in range(self.attn_heads):
            # Layer kernel
            kernel = self.add_weight(shape=(input_dim, self.units),
                                     initializer=self.kernel_initializer,
                                     regularizer=self.kernel_regularizer,
                                     constraint=self.kernel_constraint,
                                     name=f'kernel_{head}')
            self.kernels.append(kernel)

            # Layer bias
            if self.use_bias:
                bias = self.add_weight(shape=(self.units, ),
                                       initializer=self.bias_initializer,
                                       regularizer=self.bias_regularizer,
                                       constraint=self.bias_constraint,
                                       name=f'bias_{head}')
                self.biases.append(bias)

            # Attention kernels
            attn_kernel_self = self.add_weight(shape=(self.units, 1),
                                               initializer=self.attn_kernel_initializer,
                                               regularizer=self.attn_kernel_regularizer,
                                               constraint=self.attn_kernel_constraint,
                                               name=f'attn_kernel_self_{head}')
            attn_kernel_neighs = self.add_weight(shape=(self.units, 1),
                                                 initializer=self.attn_kernel_initializer,
                                                 regularizer=self.attn_kernel_regularizer,
                                                 constraint=self.attn_kernel_constraint,
                                                 name=f'attn_kernel_neigh_{head}')
            self.attn_kernels.append([attn_kernel_self, attn_kernel_neighs])

        super().build(input_shape)

    def call(self, inputs):
        '''
        inputs: (x, adj), x is node feature matrix with shape [N, F], 
        adj is adjacency matrix with shape [N, N].

        Note:
        N: number of nodes
        F: input dim
        F': output dim
        '''
        x, adj = inputs

        outputs = []
        for head in range(self.attn_heads):
            kernel = self.kernels[head]  # W in the paper
            attn_kernel_self, attn_kernel_neighs = self.attn_kernels[head]  # Attention kernel a in the paper (2F' x 1)

            # Compute inputs to attention network
            h = x @ kernel  # [N, F']

            # Compute attentions for self and neighbors
            attn_for_self = h @ attn_kernel_self  # [N, 1]
            attn_for_neighs = h @ attn_kernel_neighs  # [N, 1]

            # combine the attention with adjacency matrix via broadcast
            attn_for_self = adj * attn_for_self
            attn_for_neighs = adj * tf.transpose(attn_for_neighs)

            # Attention head a(Wh_i, Wh_j) = a^T [[Wh_i], [Wh_j]]
            attentions = tf.sparse.add(attn_for_self, attn_for_neighs)

            # Add nonlinearty by LeakyReLU
            attentions = tf.sparse.SparseTensor(indices=attentions.indices,
                                                values=LeakyReLU(alpha=0.2)(attentions.values),
                                                dense_shape=attentions.dense_shape
                                                )
            # Apply softmax to get attention coefficients
            attentions = tf.sparse.softmax(attentions)  # (N x N)

            # Apply dropout to features and attention coefficients
            if self.dropout is not None:
                attentions = tf.sparse.SparseTensor(indices=attentions.indices,
                                                    values=Dropout(rate=self.dropout)(attentions.values),
                                                    dense_shape=attentions.dense_shape
                                                    )  # (N x N)
                h = Dropout(self.dropout)(h)  # (N x F')

            # Linear combination with neighbors' features
            h = tf.sparse.sparse_dense_matmul(attentions, h)  # (N x F')

            if self.use_bias:
                h += self.biases[head]

            # Add output of attention head to final output
            outputs.append(h)

        # Aggregate the heads' output according to the reduction method
        if self.attn_heads_reduction == 'concat':
            output = tf.concat(outputs, axis=1)  # (N x KF')
        else:
            output = tf.reduce_mean(tf.stack(outputs), axis=0)  # (N x F')

        return self.activation(output)

    def get_config(self):

        config = {'units': self.units,
                  'attn_heads': self.attn_heads,
                  'attn_heads_reduction': self.attn_heads_reduction,
                  'use_bias': self.use_bias,
                  'activation': keras.activations.serialize(self.activation),
                  'kernel_initializer': keras.initializers.serialize(
                      self.kernel_initializer),
                  'attn_kernel_initializer': keras.initializers.serialize(
                      self.attn_kernel_initializer),
                  'bias_initializer': keras.initializers.serialize(
                      self.bias_initializer),
                  'kernel_regularizer': keras.regularizers.serialize(
                      self.kernel_regularizer),
                  'attn_kernel_constraint': keras.regularizers.serialize(
                      self.attn_kernel_constraint),
                  'bias_regularizer': keras.regularizers.serialize(
                      self.bias_regularizer),
                  'activity_regularizer': keras.regularizers.serialize(
                      self.activity_regularizer),
                  'kernel_constraint': keras.constraints.serialize(
                      self.kernel_constraint),
                  'attn_kernel_constraint': keras.constraints.serialize(
                      self.attn_kernel_constraint),
                  'bias_constraint': keras.constraints.serialize(self.bias_constraint)
                  }

        base_config = super().get_config()
        return {**base_config, **config}

    def compute_output_shape(self, input_shape):
        output_shape = input_shape[0][0], self.output_dim
        return tf.TensorShape(output_shape)



import numpy as np
import tensorflow as tf
import scipy.sparse as sp

from numbers import Number
from tensorflow.keras import backend as K

from graphgallery import config
from graphgallery.utils.type_check import is_list_like, is_interger_scalar, is_tensor_or_variable


__all__ = ['chk_convert', 'sparse_adj_to_sparse_tensor', 'sparse_tensor_to_sparse_adj',
           'sparse_adj_to_edges', 'edges_to_sparse_adj', 'asintarr', 'astensor',
           ]


def chk_convert(matrix, is_sparse):
    """Check the input matrix and convert it into a proper data type.

    Arguments:
    ----------
        matrix: Scipy sparse matrix or Numpy array-like or Numpy matrix. 
        is_sparse: Indicating whether the input matrix is sparse matrix or not.

    Returns:
    ----------
        A converted matrix with proper data type.

    """
    if not is_sparse:
        if not isinstance(matrix, (np.ndarray, np.matrix)):
            raise TypeError("The input matrix must be Numpy array-like or `np.matrix`"
                            f" when `is_sparse=False`, but got {type(x)}")
        return np.asarray(matrix, dtype=config.floatx())
    else:
        if not sp.issparse(matrix):
            raise TypeError(f"The input matrix must be Scipy sparse matrix when `is_sparse=True`, but got {type(x)}")

        return matrix.astype(dtype=config.floatx(), copy=False)


def sparse_adj_to_sparse_tensor(x):
    """Converts a Scipy sparse matrix to a SparseTensor."""
    sparse_coo = x.tocoo()
    row, col = sparse_coo.row, sparse_coo.col
    data, shape = sparse_coo.data, sparse_coo.shape
    if issubclass(data.dtype.type, np.floating):
        data = data.astype(config.floatx())
    indices = np.concatenate(
        (np.expand_dims(row, axis=1), np.expand_dims(col, axis=1)), axis=1)
    return tf.sparse.SparseTensor(indices, data, shape)


def sparse_tensor_to_sparse_adj(x):
    """Converts a SparseTensor to a Scipy sparse matrix."""
    # TODO
    return x


def sparse_adj_to_edges(adj):
    """Convert a Scipy sparse matrix to (edge_index, edge_weight) representation

    edge_index: shape [M, 2]
    edge_weight: shape [M,]

    """
    adj = adj.tocoo()
    edge_index = np.stack([adj.row, adj.col], axis=1).astype(config.intx())
    edge_weight = adj.data.astype(config.floatx())

    return edge_index, edge_weight


def edges_to_sparse_adj(edge_index, edge_weight):
    """Convert (edge_index, edge_weight) representation to a Scipy sparse matrix

    edge_index: shape [M, 2]
    edge_weight: shape [M,]

    """
    n = np.max(edge_index) + 1
    adj = sp.csr_matrix((edge_weight, (edge_index[:, 0], edge_index[:, 1])), shape=(n, n))
    return adj.astype(config.floatx(), copy=False)


def inferer_type(x):
    x = np.asarray(x)
    if x.dtype.kind == 'f':
        return config.floatx()
    elif x.dtype.kind == 'i':
        return config.intx()
    elif x.dtype.kind == 'b':
        return 'bool'
    else:
        raise RuntimeError(f'Invalid types, type `{type(x)}`')


def asintarr(matrix, dtype=config.intx()):
    """Convert `matrix` to interger data type.

    """
    if is_tensor_or_variable(matrix):
        return tf.cast(matrix, dtype=dtype)

    if is_interger_scalar(matrix):
        matrix = np.asarray([matrix], dtype=dtype)
    elif is_list_like(matrix) or isinstance(matrix, (np.ndarray, np.matrix)):
        matrix = np.asarray(matrix, dtype=dtype)
    else:
        raise TypeError(f'Invalid input matrix which should be either array-like or integer scalar, but got {type(matrix)}.')
    return matrix


def astensor(inputs):
    """Convert input matrices to Tensors (SparseTensors).

    inputs: single or a list of array-like variables.
    """
    def matrix_astensor(matrix):
        if any((is_tensor_or_variable(matrix), K.is_sparse(matrix), matrix is None)):
            return matrix
        elif sp.isspmatrix_csr(matrix) or sp.isspmatrix_csc(matrix):
            return sparse_adj_to_sparse_tensor(matrix)
        elif isinstance(matrix, (np.ndarray, np.matrix)) or is_list_like(matrix):
            return tf.convert_to_tensor(matrix, dtype=inferer_type(matrix))
        else:
            raise TypeError(f'Invalid type `{type(matrix)}` of inputs data. Allowed data type `(Tensor, SparseTensor, numpy array, scipy sparse tensor, None)`, but got {type(matrix)}.')

    # Check `not isinstance(inputs[0], Number)` to avoid the situation like [1,2,3],
    # where [1,2,3] will be converted to three tensors seperately.
    if is_list_like(inputs) and not isinstance(inputs[0], Number):
        return [astensor(matrix) for matrix in inputs]
    else:
        return matrix_astensor(inputs)



import tensorflow as tf
from tensorflow.keras import activations, constraints, initializers, regularizers
from tensorflow.keras.layers import Layer, Dropout, Conv1D


class LGConvolution(Layer):
    """
        Large-scale graph convolution layer as in:
        [Large-Scale Learnable Graph Convolutional Networks](https://arxiv.org/abs/1808.03965)
        Tensorflow 1.x implementation: https://github.com/divelab/lgcn

        `LGConvolution` implements the operation:
        `output = Conv1d(Conv1d(x))`, where `x` is the input feature matrix, 
        and the dropout will be used in `x` and hidden outpus.


        Arguments:
            filters: Integer, related to the dimensionality of the output space
              (i.e. the number of output filters in the convolution).
            kernel_size: An integer, related to the
              height and width of the 2D convolution window.
              Can be a single integer to specify the same value for
              all spatial dimensions.              
            dropout: Float, the dropout rate of inputs and hidden outputs.
            use_bias: Boolean, whether the layer uses a bias vector.
            activation: Activation function to use.
              If you don't specify anything, no activation is applied
              (ie. "linear" activation: `a(x) = x`).
            kernel_initializer: Initializer for the `kernel` weights matrix.
            bias_initializer: Initializer for the bias vector.
            kernel_regularizer: Regularizer function applied to
              the `kernel` weights matrix.
            bias_regularizer: Regularizer function applied to the bias vector.
            activity_regularizer: Regularizer function applied to
              the output of the layer (its "activation")..
            kernel_constraint: Constraint function applied to
              the `kernel` weights matrix.
            bias_constraint: Constraint function applied to the bias vector.

        Input shape:
            2-D tensor: Tensor `x`: `(n_nodes, n_features)`. where `x` is the node feature matrix (Tensor).

        Output shape:
            2-D tensor with shape: `(n_nodes, filters)`.
    """

    def __init__(self, filters, kernel_size,
                 use_bias=False,
                 dropout=0.5,
                 activation=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):

        super().__init__(**kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.use_bias = use_bias
        self.dropout_rate = dropout

        self.activation = activations.get(activation)
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

    def build(self, input_shapes):

        kernel_size = self.kernel_size
        filters = self.filters
        input_dim = input_shapes[-1]

        self.dropout = Dropout(rate=self.dropout_rate)
        self.conv1 = Conv1D((input_dim+filters)//2, (kernel_size+1)//2+1,
                            use_bias=True,
                            activation=self.activation,
                            kernel_initializer=self.kernel_initializer,
                            bias_initializer=self.bias_initializer,
                            kernel_regularizer=self.kernel_regularizer,
                            bias_regularizer=self.bias_regularizer,
                            kernel_constraint=self.kernel_constraint,
                            bias_constraint=self.bias_constraint,
                            name='conv1')
        self.conv2 = Conv1D(filters, kernel_size//2+1,
                            use_bias=self.use_bias,
                            activation=self.activation,
                            kernel_initializer=self.kernel_initializer,
                            bias_initializer=self.bias_initializer,
                            kernel_regularizer=self.kernel_regularizer,
                            bias_regularizer=self.bias_regularizer,
                            kernel_constraint=self.kernel_constraint,
                            bias_constraint=self.bias_constraint,
                            name='conv2')

        super().build(input_shapes)

    def call(self, inputs):

        h = inputs
        h = self.dropout(h)
        h = self.conv1(h)
        h = self.dropout(h)
        h = self.conv2(h)
        h = tf.squeeze(h, axis=1)
        return h

    def get_config(self):
        config = {'filters': self.filters,
                  'kernel_size': self.kernel_size,
                  'use_bias': self.use_bias,
                  'activation': activations.serialize(self.activation),
                  'kernel_initializer': initializers.serialize(
                      self.kernel_initializer),
                  'bias_initializer': initializers.serialize(
                      self.bias_initializer),
                  'kernel_regularizer': regularizers.serialize(
                      self.kernel_regularizer),
                  'bias_regularizer': regularizers.serialize(
                      self.bias_regularizer),
                  'activity_regularizer': regularizers.serialize(
                      self.activity_regularizer),
                  'kernel_constraint': constraints.serialize(
                      self.kernel_constraint),
                  'bias_constraint': constraints.serialize(self.bias_constraint)
                  }

        base_config = super().get_config()
        return {**base_config, **config}

    def compute_output_shape(self, input_shapes):
        output_shape = (input_shapes[0], self.filters)
        return tf.TensorShape(output_shape)  # (n_nodes, output_dim)



import tensorflow as tf
from tensorflow.keras import Model, Input
from tensorflow.keras.layers import Dense, Dropout
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import regularizers

from graphgallery.nn.layers import GraphConvolution
from graphgallery.nn.models import SupervisedModel
from graphgallery.sequence import FastGCNBatchSequence
from graphgallery import astensor, asintarr, normalize_x, normalize_adj


class FastGCN(SupervisedModel):
    """
        Implementation of Fast Graph Convolutional Networks (FastGCN). 
        `FastGCN: Fast Learning with Graph Convolutional Networks via Importance Sampling <https://arxiv.org/abs/1801.10247>`
        Tensorflow 1.x implementation: <https://github.com/matenure/FastGCN>

        Arguments:
        ----------
            adj: shape (N, N), Scipy sparse matrix if  `is_adj_sparse=True`, 
                Numpy array-like (or matrix) if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            x: shape (N, F), Scipy sparse matrix if `is_x_sparse=True`, 
                Numpy array-like (or matrix) if `is_x_sparse=False`.
                The input node feature matrix, where `F` is the dimension of features.
            labels: Numpy array-like with shape (N,)
                The ground-truth labels for all nodes in graph.
            norm_adj_rate (Float scalar, optional): 
                The normalize rate for adjacency matrix `adj`. (default: :obj:`-0.5`, 
                i.e., math:: \hat{A} = D^{-\frac{1}{2}} A D^{-\frac{1}{2}}) 
            norm_x_type (String, optional): 
                How to normalize the node feature matrix. See `graphgallery.normalize_x`
                (default :obj: `None`, i.e., do not enforce normalize)
            batch_size (Positive integer, optional): 
                Batch size for the training nodes. (default :int: `256`)
            rank (Positive integer, optional): 
                The selected nodes for each batch nodes, `rank` must be smaller than 
                `batch_size`. (default :int: `100`)
            device (String, optional): 
                The device where the model is running on. You can specified `CPU` or `GPU` 
                for the model. (default: :str: `CPU:0`, i.e., running on the 0-th `CPU`)
            seed (Positive integer, optional): 
                Used in combination with `tf.random.set_seed` & `np.random.seed` & `random.seed`  
                to create a reproducible sequence of tensors across multiple calls. 
                (default :obj: `None`, i.e., using random seed)
            name (String, optional): 
                Specified name for the model. (default: :str: `class.__name__`)


    """

    def __init__(self, adj, x, labels, norm_adj_rate=-0.5, norm_x_type=None,
                 batch_size=256, rank=100, device='CPU:0', seed=None, name=None, **kwargs):

        super().__init__(adj, x, labels, device=device, seed=seed, name=name, **kwargs)

        self.rank = rank
        self.batch_size = batch_size
        self.norm_adj_rate = norm_adj_rate
        self.norm_x_type = norm_x_type
        self.preprocess(adj, x)

    def preprocess(self, adj, x):
        super().preprocess(adj, x)
        # check the input adj and x, and convert them into proper data types
        adj, x = self._check_inputs(adj, x)

        if self.norm_adj_rate:
            adj = normalize_adj(adj, self.norm_adj_rate)

        if self.norm_x_type:
            x = normalize_x(x, norm=self.norm_x_type)

        x = adj.dot(x)

        with tf.device(self.device):
            self.x_norm, self.adj_norm = astensor(x), adj

    def build(self, hiddens=[32], activations=['relu'], dropout=0.5,
              lr=0.01, l2_norm=5e-4, use_bias=False):

        assert len(hiddens) == len(activations), "The number of hidden units and " \
            "activation functions should be the same."
        with tf.device(self.device):

            x = Input(batch_shape=[None, self.n_features], dtype=self.floatx, name='features')
            adj = Input(batch_shape=[None, None], dtype=self.floatx, sparse=True, name='adj_matrix')

            h = x
            for hid, activation in zip(hiddens, activations):
                h = Dense(hid, use_bias=use_bias, activation=activation, kernel_regularizer=regularizers.l2(l2_norm))(h)
                h = Dropout(rate=dropout)(h)

            output = GraphConvolution(self.n_classes, activation='softmax')([h, adj])

            model = Model(inputs=[x, adj], outputs=output)
            model.compile(loss='sparse_categorical_crossentropy', optimizer=Adam(lr=lr), metrics=['accuracy'])

            self.set_model(model)

    def predict(self, index):
        super().predict(index)
        index = asintarr(index)
        adj = self.adj_norm[index]
        with tf.device(self.device):
            adj = astensor(adj)
            logit = self.model.predict_on_batch([self.x_norm, adj])

        if tf.is_tensor(logit):
            logit = logit.numpy()
        return logit

    def train_sequence(self, index):
        index = asintarr(index)
        labels = self.labels[index]
        adj = self.adj[index].tocsc(copy=False)[:, index]

        if self.norm_adj_rate:
            adj = normalize_adj(adj, self.norm_adj_rate)

        with tf.device(self.device):
            x = tf.gather(self.x_norm, index)
            sequence = FastGCNBatchSequence([x, adj], labels,
                                            batch_size=self.batch_size,
                                            rank=self.rank)
        return sequence

    def test_sequence(self, index):
        index = asintarr(index)
        labels = self.labels[index]
        adj = self.adj_norm[index]

        with tf.device(self.device):
            sequence = FastGCNBatchSequence([self.x_norm, adj],
                                            labels, batch_size=None, rank=None)  # use full batch
        return sequence



import tensorflow as tf
from tensorflow.keras import Model, Input
from tensorflow.keras.layers import Dropout, Softmax
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import regularizers
from tensorflow.keras.initializers import TruncatedNormal

from graphgallery.nn.layers import GraphConvolution
from graphgallery.nn.models import SupervisedModel
from graphgallery.sequence import FullBatchNodeSequence
from graphgallery.utils.bvat_utils import kl_divergence_with_logit, entropy_y_x
from graphgallery import astensor, asintarr, normalize_x, normalize_adj


class OBVAT(SupervisedModel):
    """
        Implementation of optimization-based Batch Virtual Adversarial Training  Graph Convolutional Networks (OBVAT). 
        `Batch Virtual Adversarial Training for Graph Convolutional Networks <https://arxiv.org/abs/1902.09192>`
        Tensorflow 1.x implementation: <https://github.com/thudzj/BVAT>

        Arguments:
        ----------
            adj: shape (N, N), Scipy sparse matrix if  `is_adj_sparse=True`, 
                Numpy array-like (or matrix) if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            x: shape (N, F), Scipy sparse matrix if `is_x_sparse=True`, 
                Numpy array-like (or matrix) if `is_x_sparse=False`.
                The input node feature matrix, where `F` is the dimension of features.
            labels: Numpy array-like with shape (N,)
                The ground-truth labels for all nodes in graph.
            norm_adj_rate (Float scalar, optional): 
                The normalize rate for adjacency matrix `adj`. (default: :obj:`-0.5`, 
                i.e., math:: \hat{A} = D^{-\frac{1}{2}} A D^{-\frac{1}{2}}) 
            norm_x_type (String, optional): 
                How to normalize the node feature matrix. See `graphgallery.normalize_x`
                (default :str: `l1`)
            device (String, optional): 
                The device where the model is running on. You can specified `CPU` or `GPU` 
                for the model. (default: :str: `CPU:0`, i.e., running on the 0-th `CPU`)
            seed (Positive integer, optional): 
                Used in combination with `tf.random.set_seed` & `np.random.seed` & `random.seed`  
                to create a reproducible sequence of tensors across multiple calls. 
                (default :obj: `None`, i.e., using random seed)
            name (String, optional): 
                Specified name for the model. (default: :str: `class.__name__`)


    """

    def __init__(self, adj, x, labels, norm_adj_rate=-0.5, norm_x_type='l1',
                 device='CPU:0', seed=None, name=None, **kwargs):

        super().__init__(adj, x, labels, device=device, seed=seed, name=name, **kwargs)

        self.norm_adj_rate = norm_adj_rate
        self.norm_x_type = norm_x_type
        self.preprocess(adj, x)
        self.do_before_train = self.extra_train

    def preprocess(self, adj, x):
        super().preprocess(adj, x)
        # check the input adj and x, and convert them into proper data types
        adj, x = self._check_inputs(adj, x)

        if self.norm_adj_rate:
            adj = normalize_adj(adj, self.norm_adj_rate)

        if self.norm_x_type:
            x = normalize_x(x, norm=self.norm_x_type)

        with tf.device(self.device):
            self.x_norm, self.adj_norm = astensor([x, adj])

    def build(self, hiddens=[16], activations=['relu'], dropout=0.5, lr=0.01, l2_norm=5e-4, p1=1.4, p2=0.7, ensure_shape=True):

        with tf.device(self.device):

            x = Input(batch_shape=[None, self.n_features], dtype=self.floatx, name='features')
            adj = Input(batch_shape=[None, None], dtype=self.floatx, sparse=True, name='adj_matrix')
            index = Input(batch_shape=[None],  dtype=self.intx, name='index')

            self.GCN_layers = [GraphConvolution(hiddens[0], activation=activations[0],
                                                kernel_regularizer=regularizers.l2(l2_norm)),
                               GraphConvolution(self.n_classes)]
            self.dropout_layer = Dropout(rate=dropout)

            logit = self.propagation(x, adj)
            # To aviod the UserWarning of `tf.gather`, but it causes the shape
            # of the input data to remain the same
            if ensure_shape:
                logit = tf.ensure_shape(logit, (self.n_nodes, self.n_classes))
            output = tf.gather(logit, index)
            output = Softmax()(output)
            model = Model(inputs=[x, adj, index], outputs=output)
            model.compile(loss='sparse_categorical_crossentropy', optimizer=Adam(lr=lr), metrics=['accuracy'])

            self.r_vadv = tf.Variable(TruncatedNormal(stddev=0.01)(shape=[self.n_nodes, self.n_features]), name="r_vadv")
            entropy_loss = entropy_y_x(logit)
            vat_loss = self.virtual_adversarial_loss(x, adj, logit)
            model.add_loss(p1 * vat_loss + p2 * entropy_loss)

            self.set_model(model)
            self.adv_optimizer = Adam(lr=lr/10)

    def virtual_adversarial_loss(self, x, adj, logit):

        adv_x = x + self.r_vadv
        logit_p = tf.stop_gradient(logit)
        logit_m = self.propagation(adv_x, adj)
        loss = kl_divergence_with_logit(logit_p, logit_m)
        return loss

    def propagation(self, x, adj):
        h = x
        for layer in self.GCN_layers:
            h = self.dropout_layer(h)
            h = layer([h, adj])
        return h

    @tf.function
    def extra_train(self, epochs=10):

        with tf.device(self.device):
            r_vadv = self.r_vadv
            optimizer = self.adv_optimizer
            x, adj = self.x_norm, self.adj_norm
            r_vadv.assign(TruncatedNormal(stddev=0.01)(shape=tf.shape(r_vadv)))
            for _ in range(epochs):
                with tf.GradientTape() as tape:
                    rnorm = tf.nn.l2_loss(r_vadv)
                    logit = self.propagation(x, adj)
                    vloss = self.virtual_adversarial_loss(x, adj, logit)
                    loss = rnorm - vloss
                gradient = tape.gradient(loss, r_vadv)
                optimizer.apply_gradients(zip([gradient], [r_vadv]))

    def predict(self, index):
        super().predict(index)
        index = asintarr(index)

        with tf.device(self.device):
            index = astensor(index)
            logit = self.model.predict_on_batch([self.x_norm, self.adj_norm, index])

        if tf.is_tensor(logit):
            logit = logit.numpy()
        return logit

    def train_sequence(self, index):
        index = asintarr(index)
        labels = self.labels[index]

        with tf.device(self.device):
            sequence = FullBatchNodeSequence([self.x_norm, self.adj_norm, index], labels)

        return sequence



import os
import random
import logging

import numpy as np
import tensorflow as tf
import scipy.sparse as sp

from graphgallery import config, chk_convert, asintarr


class BaseModel:
    """Base model for supervised learning and unsupervised learning.


    Arguments:
    ----------
        adj: shape (N, N), `scipy.sparse.csr_matrix` (or `csc_matrix`) if 
            `is_adj_sparse=True`, Numpy array-like or `np.matrix` if `is_adj_sparse=False`.
            The input `symmetric` adjacency matrix, where `N` is the number 
            of nodes in graph.
        x: shape (N, F), `scipy.sparse.csr_matrix` (or `csc_matrix`) if 
            `is_x_sparse=True`, Numpy array-like or `np.matrix` if `is_x_sparse=False`.
            The input node feature matrix, where `F` is the dimension of features.
        labels: shape (N,), array-like. Default: `None` for unsupervised learning.
            The class labels of the nodes in the graph. 
        device: string. Default: `CPU:0`.
            The device where the model running on.
        seed: interger scalar. Default: `None`.
            Used in combination with `tf.random.set_seed` & `np.random.seed` & `random.seed`  
            to create a reproducible sequence of tensors across multiple calls. 
        name (String, optional): 
                Specified name for the model. (default: :str: `class.__name__`)
        kwargs: other customed keyword arguments.

    Note:
    ----------
        By default, `adj` is sparse matrix and `x` is dense array. Both of them are 
        2-D matrices.

    """

    def __init__(self, adj, x, labels=None, device="CPU:0", seed=None, name=None, **kwargs):

        if seed:
            tf.random.set_seed(seed)
            np.random.seed(seed)
            random.seed(seed)

        if name is None:
            name = self.__class__.__name__

        # By default, adj is sparse matrix and x is dense array (matrix)
        is_adj_sparse = kwargs.pop("is_adj_sparse", True)
        is_x_sparse = kwargs.pop("is_x_sparse", False)
        is_weighted = kwargs.pop("is_weighted", False)

        # leave it blank for future
        allowed_kwargs = set([])
        unknown_kwargs = set(kwargs.keys()) - allowed_kwargs
        if unknown_kwargs:
            raise ValueError(
                "Invalid keyword argument(s) in `__init__`: %s" % (unknown_kwargs,))

        # check the input adj and x, and convert them to appropriate forms
        self.n_nodes, self.n_features = x.shape

        self.is_adj_sparse = is_adj_sparse
        self.is_x_sparse = is_x_sparse
        self.adj, self.x = self._check_inputs(adj, x)   

        if labels is not None:
            self.n_classes = np.max(labels) + 1
        else:
            self.n_classes = None

        self.seed = seed
        self.device = device
        self.labels = asintarr(labels)
        self.name = name
        self._model = None
        self.index_train = None
        self.index_val = None
        self.index_test = None
        self.do_before_train = None
        self.do_before_validation = None
        self.do_before_test = None
        self.do_before_predict = None
        self.sparse = True
        self.norm_x_fn = None
        self.custom_objects = None  # used for save/load model

        self.weight_path = f"./weight/{name}_weights"

        # data types, default: `float32` and `int64`
        self.floatx = config.floatx()
        self.intx = config.intx()

    def _check_inputs(self, adj, x):
        """Check the input adj and x and make sure they are legal forms of input.

        Arguments:
        ----------
            adj: shape (N, N), `scipy.sparse.csr_matrix` (or `csc_matrix`) if 
                `is_adj_sparse=True`, Numpy array-like or `np.matrix` if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            x: shape (N, F), `scipy.sparse.csr_matrix` (or `csc_matrix`) if 
                `is_x_sparse=True`, Numpy array-like or `np.matrix` if `is_x_sparse=False`.
                The input node feature matrix, where `F` is the dimension of features.

        Note:
        ----------
            By default, `adj` is sparse matrix and `x` is dense array. Both of them are 
            2-D matrices.


        """
        adj = chk_convert(adj, self.is_adj_sparse)
        x = chk_convert(x, self.is_x_sparse)
        adj_shape = adj.shape
        x_shape = x.shape

        if adj_shape[0] != x_shape[0]:
            raise RuntimeError(f"The first dimension of adjacency matrix and feature matrix should be equal.")

        if len(adj_shape) != len(x_shape) != 2:
            raise RuntimeError(f"The adjacency matrix and feature matrix should have the SAME dimensions 2.")

        if adj_shape[0] != adj_shape[1]:
            raise RuntimeError(f"The adjacency matrix should be N by N square matrix.")
        return adj, x

    @property
    def model(self):
        return self._model

    def set_model(self, model):
        self._model = model

    def preprocess(self, adj, x):
        """Preprocess the input adjacency matrix and feature matrix, e.g., normalization.
        And convert them to tf.tensor. 

        Arguments:
        ----------
            adj: shape (N, N), `scipy.sparse.csr_matrix` (or `csc_matrix`) if 
                `is_adj_sparse=True`, Numpy array-like if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            x: shape (N, F), `scipy.sparse.csr_matrix` (or `csc_matrix`) if 
                `is_x_sparse=True`, Numpy array-like if `is_x_sparse=False`.
                The input node feature matrix, where `F` is the dimension of node features.

        Note:
        ----------
            By default, `adj` is sparse matrix and `x` is dense array. Both of them are 
            2-D matrices.
        """
        # check the input adj and x, and convert them to appropriate forms
        self.adj, self.x = self._check_inputs(adj, x)        
        self.n_nodes, self.n_features = x.shape

    def save(self, path=None, as_model=False):
        if not os.path.exists("weight"):
            os.makedirs("weight")
            logging.log(logging.WARNING, "Mkdir /weight")

        if path is None:
            path = self.weight_path

        if not path.endswith('.h5'):
            path += '.h5'

        if as_model:
            self.model.save(path, save_format="h5")
        else:
            try:
                self.model.save_weights(path)
            except ValueError as e:
                # due to the bugs in tf 2.1
                self.model.save_weights(path[:-3])

    def load(self, path=None, as_model=False):
        if path is None:
            path = self.weight_path
        if not path.endswith('.h5'):
            path += '.h5'
        if as_model:
            model = tf.keras.models.load_model(path, custom_objects=self.custom_objects)
            self.set_model(model)
        else:
            try:
                self.model.load_weights(path)
            except KeyError as e:
                self.model.load_weights(path[:-3])

    def __repr__(self):
        return f"Graphgallery.nn.{self.name} in {self.device}"



import numpy as np

from numba import njit
from gensim.models import Word2Vec

from graphgallery.nn.models import UnsupervisedModel


class Deepwalk(UnsupervisedModel):
    """
        Implementation of DeepWalk Unsupervised Graph Neural Networks (DeepWalk). 
        [DeepWalk: Online Learning of Social Representations](https://arxiv.org/abs/1403.6652)
        Implementation: https://github.com/phanein/deepwalk

        Arguments:
        ----------
            adj: shape (N, N), Scipy sparse matrix if  `is_adj_sparse=True`, 
                Numpy array-like (or matrix) if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            x: shape (N, F), Scipy sparse matrix if `is_x_sparse=True`, 
                Numpy array-like (or matrix) if `is_x_sparse=False`.
                The input node feature matrix, where `F` is the dimension of features.
            labels: Numpy array-like with shape (N,)
                The ground-truth labels for all nodes in graph.
            device (String, optional): 
                The device where the model is running on. You can specified `CPU` or `GPU` 
                for the model. (default: :str: `CPU:0`, i.e., running on the 0-th `CPU`)
            seed (Positive integer, optional): 
                Used in combination with `tf.random.set_seed` & `np.random.seed` & `random.seed`  
                to create a reproducible sequence of tensors across multiple calls. 
                (default :obj: `None`, i.e., using random seed)
            name (String, optional): 
                Specified name for the model. (default: :str: `class.__name__`)

    """

    def __init__(self, adj, x, labels, device='CPU:0', seed=None, name=None, **kwargs):

        super().__init__(adj, x, labels, device=device, seed=seed, name=name, **kwargs)

    def build(self, walk_length=80, walks_per_node=10,
              embedding_dim=64, window_size=5, workers=16,
              iter=1, num_neg_samples=1):

        walks = self.deepwalk_random_walk(self.adj.indices,
                                          self.adj.indptr,
                                          walk_length=walk_length,
                                          walks_per_node=walks_per_node)

        sentences = [list(map(str, walk)) for walk in walks]

        model = Word2Vec(sentences, size=embedding_dim, window=window_size, min_count=0, sg=1, workers=workers,
                         iter=iter, negative=num_neg_samples, hs=0, compute_loss=True)

        self.set_model(model)

    @staticmethod
    @njit
    def deepwalk_random_walk(indices, indptr, walk_length=80, walks_per_node=10):

        N = len(indptr) - 1

        for _ in range(walks_per_node):
            for n in range(N):
                single_walk = [n]
                current_node = n
                for _ in range(walk_length-1):
                    neighbors = indices[indptr[current_node]:indptr[current_node + 1]]
                    if neighbors.size == 0:
                        break
                    current_node = np.random.choice(neighbors)
                    single_walk.append(current_node)

                yield single_walk

    def get_embeddings(self, norm=True):
        embeddings = self.model.wv.vectors[np.fromiter(map(int, self.model.wv.index2word), np.int32).argsort()]

        if norm:
            embeddings = self.normalize_embedding(embeddings)

        self.embeddings = embeddings



import random
import numpy as np

def create_alias_table(area_ratio):
    """

    :param area_ratio: sum(area_ratio)=1
    :return: accept,alias
    """
    l = len(area_ratio)
    accept, alias = [0] * l, [0] * l
    small, large = [], []
    area_ratio_ = np.array(area_ratio) * l
    for i, prob in enumerate(area_ratio_):
        if prob < 1.0:
            small.append(i)
        else:
            large.append(i)

    while small and large:
        small_idx, large_idx = small.pop(), large.pop()
        accept[small_idx] = area_ratio_[small_idx]
        alias[small_idx] = large_idx
        area_ratio_[large_idx] = area_ratio_[large_idx] - (1 - area_ratio_[small_idx])
        if area_ratio_[large_idx] < 1.0:
            small.append(large_idx)
        else:
            large.append(large_idx)

    while large:
        large_idx = large.pop()
        accept[large_idx] = 1
    while small:
        small_idx = small.pop()
        accept[small_idx] = 1

    return accept, alias


def alias_sample(accept, alias):
    """

    :param accept:
    :param alias:
    :return: sample index
    """
    N = len(accept)
    i = int(random.random()*N)
    r = random.random()
    if r < accept[i]:
        return i
    else:
        return alias[i]

class RandomWalker:
    def __init__(self, G, p=1, q=1):
        """
        :param G:
        :param p: Return parameter,controls the likelihood of immediately revisiting a node in the walk.
        :param q: In-out parameter,allows the search to differentiate between “inward” and “outward” nodes
        """
        self.G = G
        self.p = p
        self.q = q

    def get_alias_edge(self, t, v):
        """
        compute unnormalized transition probability between nodes v and its neighbors give the previous visited node t.
        :param t:
        :param v:
        :return:
        """
        G = self.G
        p = self.p
        q = self.q

        unnormalized_probs = []
        for x in G.neighbors(v):
            weight = G[v][x].get('weight', 1.0)  # w_vx
            
            if x == t:  # d_tx == 0
                unnormalized_probs.append(weight/p)
            elif G.has_edge(x, t):  # d_tx == 1
                unnormalized_probs.append(weight)
            else:  # d_tx > 1
                unnormalized_probs.append(weight/q)
                
        norm_const = sum(unnormalized_probs)
        normalized_probs = [float(u_prob)/norm_const for u_prob in unnormalized_probs]

        return create_alias_table(normalized_probs)

    def preprocess_transition_probs(self):
        """
        Preprocessing of transition probabilities for guiding the random walks.
        """
        G = self.G

        alias_nodes = {}
        for node in G.nodes():
            unnormalized_probs = [G[node][nbr].get('weight', 1.0)
                                  for nbr in G.neighbors(node)]
            norm_const = sum(unnormalized_probs)
            normalized_probs = [float(u_prob)/norm_const for u_prob in unnormalized_probs]
            alias_nodes[node] = create_alias_table(normalized_probs)

        alias_edges = {}

        for edge in G.edges():
            alias_edges[edge] = self.get_alias_edge(edge[0], edge[1])

        self.alias_nodes = alias_nodes
        self.alias_edges = alias_edges



import tensorflow as tf
from tensorflow.keras.layers import Layer


class SqueezedSparseConversion(Layer):
    def __init__(self, n_nodes=None):
        super().__init__()
        self.n_nodes = n_nodes

    def call(self, inputs):
        indices, values = inputs
        n_nodes = self.n_nodes or tf.reduce_max(indices) + 1
        self.n_nodes = n_nodes

        if indices.dtype != tf.int64:
            indices = tf.cast(indices, tf.int64)

        # Build sparse tensor for the matrix
        output = tf.sparse.SparseTensor(
            indices=indices, values=values, dense_shape=(n_nodes, n_nodes)
        )
        return output

    def compute_output_shape(self, input_shapes):
        return tf.TensorShape([self.n_nodes, self.n_nodes])


class Scale(Layer):
    def call(self, inputs):
        output = (inputs - tf.reduce_mean(inputs, axis=0, keepdims=True)) / tf.keras.backend.std(inputs, axis=0, keepdims=True)
        return output

    def compute_output_shape(self, input_shapes):
        return tf.TensorShape(input_shapes)



import scipy.sparse as sp
import numpy as np

from sklearn.preprocessing import normalize

# Version: Compute the exact signature
# def laplacian(W, normalized=True):
#     """Return the Laplacian of the weight matrix."""
#     # Degree matrix.
#     d = W.sum(axis=0).A1
#     # Laplacian matrix.
#     if not normalized:
#         D = sp.diags(d)
#         L = D - W
#     else:
#         d = 1 / np.sqrt(d)
#         D = sp.diags(d)
#         I = sp.identity(d.size, dtype=W.dtype)
#         L = I - D * W * D

#     return L

# def fourier(L, algo='eigh', k=100):
#     """Return the Fourier basis, i.e. the EVD of the Laplacian."""
#     def sort(lamb, U):
#         idx = lamb.argsort()
#         return lamb[idx], U[:, idx]

#     if algo is 'eig':
#         lamb, U = np.linalg.eig(L.toarray())
#     elif algo is 'eigh':
#         lamb, U = np.linalg.eigh(L.toarray())
#     elif algo is 'eigs':
#         lamb, U = sp.linalg.eigs(L, k=k, which='SM')
#     elif algo is 'eigsh':
#         lamb, U = sp.linalg.eigsh(L, k=k, which='SM')
#     else:
#         raise ValueError(f'Invalid argument of {algo}.')
        
#     lamb, U = sort(lamb, U)

#     return lamb, U

# def weight_wavelet(wavelet_s, lamb, U):
#     # In original code, lamb has changed but here hasn't
#     lamb = np.exp(-lamb*wavelet_s)
#     Weight = (U * lamb) @ U.T
#     return Weight
# #     return Weight, lamb

# def weight_wavelet_inverse(wavelet_s, lamb, U):
#     lamb = np.exp(lamb*wavelet_s)
#     Weight = (U * lamb) @ U.T
#     return Weight



# def wavelet_basis(adj, wavelet_s=1.0, laplacian_normalize=True, 
#                   sparseness=True, threshold=1e-4, weight_normalize=False, k=100):

#     L = laplacian(adj, normalized=laplacian_normalize)
#     lamb, U = fourier(L, k=k)
# #     Weight, lamb = weight_wavelet(wavelet_s, lamb, U)
#     Weight = weight_wavelet(wavelet_s, lamb, U)
#     inverse_Weight = weight_wavelet_inverse(wavelet_s, lamb, U)
#     del U, lamb

#     if sparseness:
#         Weight[Weight < threshold] = 0.0
#         inverse_Weight[inverse_Weight < threshold] = 0.0

#     if weight_normalize:
#         Weight = normalize(Weight, norm='l1', axis=1)
#         inverse_Weight = normalize(inverse_Weight, norm='l1', axis=1)

#     Weight = sp.csr_matrix(Weight)
#     inverse_Weight = sp.csr_matrix(inverse_Weight)
#     t_k = Weight, inverse_Weight
#     return t_k

# ==================================================================
# Version: Approximate with Chebychev polynomial

def laplacian(W, normalized=True):
    """Return the Laplacian of the weight matrix."""
    # Degree matrix.
    d = W.sum(axis=0).A1
    # Laplacian matrix.
    if not normalized:
        D = sp.diags(d)
        L = D - W
    else:
        d = 1 / (np.sqrt(d) + 1e-6)
#         d[np.isinf(d)] = 0.
        D = sp.diags(d)
        I = sp.identity(d.size, dtype=W.dtype)
        L = I - D * W * D
    return L

def compute_cheb_coeff_basis(scale, order):
    xx = np.array([np.cos((2. * i - 1.) / (2. * order) * np.pi) for i in range(1, order + 1)])
    basis = [np.ones((1, order)), xx]
    for k in range(order + 1 - 2):
        basis.append(2* np.multiply(xx, basis[-1]) - basis[-2])
    basis = np.vstack(basis)
    f = np.exp(-scale * (xx + 1))
    products = np.einsum("j,ij->ij", f, basis)
    coeffs = 2. / order * products.sum(1)
    coeffs[0] /= 2.
    return coeffs

def wavelet_basis(adj, order=3, wavelet_s=1.0, threshold=1e-4, wavelet_normalize=False):
    lap = laplacian(adj)
    N = adj.shape[0]
    I = sp.eye(N)
    L = lap - I
    monome = {0: I, 1: L}
    
    for k in range(2, order + 1):
        monome[k] = 2 * L @ monome[k-1]- monome[k - 2]
        
    def compute_walelet(tau):
        coeffs = compute_cheb_coeff_basis(tau, order)
        w = np.sum([coeffs[k] * monome[k] for k in range(order + 1)])
        w.data = thres(w.data)
        w.eliminate_zeros()
        return w
    
    thres = np.vectorize(lambda x : x if x > threshold else 0.)
    Wavelet = compute_walelet(wavelet_s)
    Wavelet_inverse = compute_walelet(-wavelet_s)  
    
    if wavelet_normalize:
        Wavelet = normalize(Wavelet, norm='l1', axis=1)
        Wavelet_inverse = normalize(Wavelet_inverse, norm='l1', axis=1)
    
    return Wavelet, Wavelet_inverse
    


import tensorflow as tf
from tensorflow.keras.layers import Layer


class SGConvolution(Layer):
    """
        Simplifying graph convolution layer as in: 
        [Simplifying Graph Convolutional Networks](https://arxiv.org/abs/1902.07153)
        Pytorch implementation: https://github.com/Tiiiger/SGC
        
        `SGConvolution` implements the operation:
        `output = x @ adj^{order}`
        where `x` is the feature matrix, `adj` is the adjacency matrix.
        
        Note:
          This `SGConvolution` layer does not contain any trainable parameters.
        
        
        Arguments:
          order: Positive integer, the power of adjacency matrix, i.e., adj^{order}.

        Input shape:
          tuple/list with two 2-D tensor: Tensor `x` and SparseTensor `adj`: `[(n_nodes, n_features), (n_nodes, n_nodes)]`.
          The former one is the feature matrix (Tensor) and the other is adjacency matrix (SparseTensor).

        Output shape:
          2-D tensor with shape: `(n_nodes, n_features)`.       
    """
    def __init__(self, order=2, **kwargs):
        super().__init__(**kwargs)
        self.order = order
        

    def call(self, inputs):
        x, adj = inputs
        
        for _ in range(self.order):
            x = tf.sparse.sparse_dense_matmul(adj, x)

        return x
    
    
    def get_config(self):
        config = {'order': self.order}
        
        base_config = super().get_config()
        return {**base_config, **config}
    
    
    def compute_output_shape(self, input_shapes):
        features_shape = input_shapes[0]
        return tf.TensorShape(features_shape) # (n_nodes, n_features)



import tensorflow as tf


# @tf.function
def get_normalized_vector(d):
    d /= (1e-12 + tf.reduce_max(tf.abs(d)))
    d /= tf.sqrt(1e-6 + tf.reduce_sum(tf.pow(d, 2.0), 1, keepdims=True))
    return d

# @tf.function
def kl_divergence_with_logit(q_logit, p_logit, mask=None):
    q = tf.math.softmax(q_logit)
    if mask is None:
            qlogp = tf.reduce_mean(tf.reduce_sum(q * tf.math.log_softmax(p_logit), 1))
    else:
        mask /= tf.reduce_mean(mask)
        qlogp = tf.reduce_mean(tf.reduce_sum(q * tf.math.log_softmax(p_logit), 1) * mask)
        
    return -qlogp

# @tf.function
def entropy_y_x(logit):
    p = tf.math.softmax(logit)
    return -tf.reduce_mean(tf.reduce_sum(p * tf.math.log_softmax(logit), 1))


import numpy as np
import networkx as nx
import tensorflow as tf
from tensorflow.keras import Model, Input
from tensorflow.keras.layers import Dropout, Softmax
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import regularizers

from graphgallery.nn.layers import GraphConvolution
from graphgallery.nn.models import SupervisedModel
from graphgallery.sequence import ClusterMiniBatchSequence
from graphgallery.utils.graph_utils import partition_graph
from graphgallery import Bunch, sample_mask, normalize_x, normalize_adj, astensor, asintarr


class ClusterGCN(SupervisedModel):
    """
        Implementation of Cluster Graph Convolutional Networks (ClusterGCN). 

        `Cluster-GCN: An Efficient Algorithm for Training Deep and Large Graph Convolutional Networks 
        <https://arxiv.org/abs/1905.07953>`
        Tensorflow 1.x implementation: <https://github.com/google-research/google-research/tree/master/cluster_gcn>
        Pytorch implementation: <https://github.com/benedekrozemberczki/ClusterGCN>

        Arguments:
        ----------
            adj: shape (N, N), Scipy sparse matrix if  `is_adj_sparse=True`, 
                Numpy array-like (or matrix) if `is_adj_sparse=False`.
                The input `symmetric` adjacency matrix, where `N` is the number 
                of nodes in graph.
            x: shape (N, F), Scipy sparse matrix if `is_x_sparse=True`, 
                Numpy array-like (or matrix) if `is_x_sparse=False`.
                The input node feature matrix, where `F` is the dimension of features.
            labels: Numpy array-like with shape (N,)
                The ground-truth labels for all nodes in graph.
            graph (`nx.DiGraph`, optional): 
                The networkx graph which converted by `adj`, if if not specified (`None`), 
                the graph will be converted by `adj` automatically, but it will comsum lots 
                of time. (default :obj: `None`)
            n_clusters (Potitive integer): 
                The number of clusters that the graph being seperated, if not specified (`None`), 
                it will be set to the number of classes automatically. (default :obj: `None`).
            norm_adj_rate (Float scalar, optional): 
                The normalize rate for adjacency matrix `adj`. (default: :obj:`-0.5`, 
                i.e., math:: \hat{A} = D^{-\frac{1}{2}} A D^{-\frac{1}{2}})
            norm_x_type (String, optional): 
                How to normalize the node feature matrix. See `graphgallery.normalize_x`
                (default :str: `l1`)
            device (String, optional): 
                The device where the model is running on. You can specified `CPU` or `GPU` 
                for the model. (default: :str: `CPU:0`, i.e., running on the 0-th `CPU`)
            seed (Positive integer, optional): 
                Used in combination with `tf.random.set_seed` & `np.random.seed` & `random.seed`  
                to create a reproducible sequence of tensors across multiple calls. 
                (default :obj: `None`, i.e., using random seed)
            name (String, optional): 
                Specified name for the model. (default: :str: `class.__name__`)

    """

    def __init__(self, adj, x, labels, graph=None, n_clusters=None,
                 norm_adj_rate=-0.5, norm_x_type='l1', device='CPU:0', seed=None, name=None, **kwargs):

        super().__init__(adj, x, labels=labels, device=device, seed=seed, name=name, **kwargs)

        if n_clusters is None:
            n_clusters = self.n_classes

        self.n_clusters = n_clusters
        self.norm_adj_rate = norm_adj_rate
        self.norm_x_type = norm_x_type
        self.preprocess(adj, x, graph)

    def preprocess(self, adj, x, graph=None):
        super().preprocess(adj, x)
        # check the input adj and x, and convert them into proper data types
        adj, x = self._check_inputs(adj, x)

        if self.norm_x_type:
            x = normalize_x(x, norm=self.norm_x_type)

        if not graph:
            graph = nx.from_scipy_sparse_matrix(adj, create_using=nx.DiGraph)

        (self.batch_adj, self.batch_x, self.batch_labels,
         self.cluster_member, self.mapper) = partition_graph(adj, x, self.labels, graph,
                                                             n_clusters=self.n_clusters)

        if self.norm_adj_rate:
            self.batch_adj = normalize_adj(self.batch_adj, self.norm_adj_rate)

        with tf.device(self.device):
            self.batch_adj, self.batch_x = astensor([self.batch_adj, self.batch_x])

    def build(self, hiddens=[32], activations=['relu'], dropout=0.5, lr=0.01, l2_norm=1e-5):

        assert len(hiddens) == len(activations), "The number of hidden units and " \
            "activation functions should be the same."

        with tf.device(self.device):

            x = Input(batch_shape=[None, self.n_features], dtype=self.floatx, name='features')
            adj = Input(batch_shape=[None, None], dtype=self.floatx, sparse=True, name='adj_matrix')
            mask = Input(batch_shape=[None],  dtype=tf.bool, name='mask')

            h = Dropout(rate=dropout)(x)

            for hid, activation in zip(hiddens, activations):
                h = GraphConvolution(hid, activation=activation, kernel_regularizer=regularizers.l2(l2_norm))([h, adj])
                h = Dropout(rate=dropout)(h)

            h = GraphConvolution(self.n_classes)([h, adj])
            h = tf.boolean_mask(h, mask)
            output = Softmax()(h)

            model = Model(inputs=[x, adj, mask], outputs=output)

            model.compile(loss='sparse_categorical_crossentropy', optimizer=Adam(lr=lr),
                          metrics=['accuracy'], experimental_run_tf_function=False)

            self.set_model(model)

    def predict(self, index):
        super().predict(index)
        index = asintarr(index)
        mask = sample_mask(index, self.n_nodes)

        orders_dict = {idx: order for order, idx in enumerate(index)}
        batch_mask, orders = [], []
        batch_x, batch_adj = [], []
        for cluster in range(self.n_clusters):
            nodes = self.cluster_member[cluster]
            mini_mask = mask[nodes]
            batch_nodes = np.asarray(nodes)[mini_mask]
            if batch_nodes.size == 0:
                continue
            batch_x.append(self.batch_x[cluster])
            batch_adj.append(self.batch_adj[cluster])
            batch_mask.append(mini_mask)
            orders.append([orders_dict[n] for n in batch_nodes])

        batch_data = tuple(zip(batch_x, batch_adj, batch_mask))

        logit = np.zeros((index.size, self.n_classes), dtype=self.floatx)
        with tf.device(self.device):
            batch_data = astensor(batch_data)
            for order, inputs in zip(orders, batch_data):
                output = self.model.predict_on_batch(inputs)
                logit[order] = output

        return logit

    def train_sequence(self, index):
        index = asintarr(index)
        mask = sample_mask(index, self.n_nodes)
        labels = self.labels

        batch_mask, batch_labels = [], []
        batch_x, batch_adj = [], []
        for cluster in range(self.n_clusters):
            nodes = self.cluster_member[cluster]
            mini_mask = mask[nodes]
            mini_labels = labels[nodes][mini_mask]
            if mini_labels.size == 0:
                continue
            batch_x.append(self.batch_x[cluster])
            batch_adj.append(self.batch_adj[cluster])
            batch_mask.append(mini_mask)
            batch_labels.append(mini_labels)

        batch_data = tuple(zip(batch_x, batch_adj, batch_mask))
        with tf.device(self.device):
            sequence = ClusterMiniBatchSequence(batch_data, batch_labels)
        return sequence



from graphgallery.utils.conversion import (chk_convert, sparse_adj_to_edges,
                                           sparse_tensor_to_sparse_adj,
                                           sparse_adj_to_edges,
                                           edges_to_sparse_adj,
                                           asintarr,
                                           astensor)

from graphgallery.utils.data_utils import normalize_adj, normalize_x, Bunch, sample_mask
from graphgallery.utils.shape_utils import repeat

from graphgallery import nn
from graphgallery import utils
from graphgallery import sequence



__version__ = '0.1.3'

__all__ = ['graphgallery', 'nn', 'utils', 'sequence', '__version__']


#!/usr/bin/env python
# coding: utf-8
from setuptools import setup, find_packages

__version__ = '0.1.3'
url = 'https://github.com/EdisonLeeeee/Graphgallery'

install_requires = [
            'networkx==2.3',
            'metis==0.2a4',
            'scipy==1.4.1',
            'tensorflow>=2.1.0',
            'numpy==1.17.4',
            'gensim>=3.8.0',
            'texttable>=1.6.2',
            'numba>=0.46.0',
            'tqdm>=4.40.2',
            'scikit_learn>=0.23.1',
]

setup(
    name='graphgallery',
    version=__version__,
    description='Geometric Deep Learning Extension Library for TensorFlow',
    author='Jintang Li',
    author_email='cnljt@outlook.com',
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url=url,
    download_url='{}/archive/{}.tar.gz'.format(url, __version__),
    keywords=[
        'tensorflow',
        'geometric-deep-learning',
        'graph-neural-networks',
    ],
    python_requires='>=3.6',
    install_requires=install_requires,
    packages=find_packages()
)



from tensorflow.keras import activations, constraints, initializers, regularizers
from tensorflow.keras.layers import Layer

import tensorflow as tf


class MeanAggregator(Layer):
    """
        Basic graphSAGE convolution layer as in: 
        [Inductive Representation Learning on Large Graphs](https://arxiv.org/abs/1706.02216)
        Tensorflow 1.x implementation: https://github.com/williamleif/GraphSAGE
        Pytorch implementation: https://github.com/williamleif/graphsage-simple/

        Aggregates via mean followed by matmul and non-linearity.

        `MeanAggregator` implements the operation:
        `output = activation(Concat(x @ kernel_0, Agg(neigh_x) @ kernel_1) + bias)`
        where `x` is the feature matrix, `neigh_x` is the feature matrix of neighbors,
        `Agg` is the operation of aggregation (`mean`, `sum`, `max`, `min`) along the last dimension,
        `Concat` is the operation of concatenation between transformed node features and neighbor features,
        and it could be replaced with `Add` operation.
        `activation` is the element-wise activation function
        passed as the `activation` argument, `kernel` is a weights matrix
        created by the layer, and `bias` is a bias vector created by the layer
        (only applicable if `use_bias` is `True`).


        Arguments:
          units: Positive integer, dimensionality of the output space.
          concat: Boolean, whether the layer uses concatenation 
            between transformed node features and neighbor features, 
            if `False`, the `Add` operation will be used.
          use_bias: Boolean, whether the layer uses a bias vector.
          agg_method: String, which method of aggregation will be used for neighbor 
            feature matrix, one of (`mean`, `sum`, `max`, `min`) will be used.
          activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).            
          kernel_initializer: Initializer for the `kernel` weights matrix.
          bias_initializer: Initializer for the bias vector.
          kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
          bias_regularizer: Regularizer function applied to the bias vector.
          activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
          kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
          bias_constraint: Constraint function applied to the bias vector.

        Input shape:
          tuple/list with two tensor: 2-D Tensor `x` and 3-D Tensor `neigh_x`: 
          `[(batch_n_nodes, n_features), (batch_n_nodes, n_samples, n_features)]`.
          The former one is the node feature matrix (Tensor) and the last is the neighbor feature matrix (Tensor).

        Output shape:
          2-D tensor with shape: `(batch_n_nodes, units)` or `(batch_n_nodes, units * 2)`,
          depend on using `Add` or `Concat` for the node and neighbor features.       
    """

    def __init__(self,
                 units,
                 concat=False,
                 use_bias=True,
                 agg_method='mean',
                 activation=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):

        super().__init__(**kwargs)
        self.units = units
        self.concat = concat
        self.use_bias = use_bias
        self.agg_method = agg_method
        self.aggregator = {'mean': tf.reduce_mean, 'sum': tf.reduce_sum,
                           'max': tf.reduce_max, 'min': tf.reduce_min}[agg_method]
        self.activation = activations.get(activation)

        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)

        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)

        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

        if concat:
            self.output_dim = units * 2
        else:
            self.output_dim = units

    def build(self, input_shape):
        input_dim = input_shape[0][-1]

        self.kernel_self = self.add_weight(shape=(input_dim, self.units),
                                           initializer=self.kernel_initializer,
                                           regularizer=self.kernel_regularizer,
                                           constraint=self.kernel_constraint,
                                           name='kernel_self')
        self.kernel_neigh = self.add_weight(shape=(input_dim, self.units),
                                            initializer=self.kernel_initializer,
                                            regularizer=self.kernel_regularizer,
                                            constraint=self.kernel_constraint,
                                            name='kernel_neigh')

        # Layer bias
        if self.use_bias:
            self.bias = self.add_weight(shape=(self.output_dim, ),
                                        initializer=self.bias_initializer,
                                        regularizer=self.bias_regularizer,
                                        constraint=self.bias_constraint,
                                        name='bias')

        super().build(input_shape)

    def call(self, inputs):
        x, neigh_x = inputs

        neigh_x = self.aggregator(neigh_x, axis=1)

        x = x @ self.kernel_self
        neigh_x = neigh_x @ self.kernel_neigh

        if self.concat:
            output = tf.concat([x, neigh_x], axis=1)
        else:
            output = x + neigh_x

        if self.use_bias:
            output += self.bias

        return self.activation(output)

    def get_config(self):

        config = {'units': self.units,
                  'concat': self.concat,
                  'use_bias': self.use_bias,
                  'agg_method': self.agg_method,
                  'activation': keras.activations.serialize(self.activation),
                  'kernel_initializer': keras.initializers.serialize(
                      self.kernel_initializer),
                  'bias_initializer': keras.initializers.serialize(
                      self.bias_initializer),
                  'kernel_regularizer': keras.regularizers.serialize(
                      self.kernel_regularizer),
                  'bias_regularizer': keras.regularizers.serialize(
                      self.bias_regularizer),
                  'activity_regularizer': keras.regularizers.serialize(
                      self.activity_regularizer),
                  'kernel_constraint': keras.constraints.serialize(
                      self.kernel_constraint),
                  'bias_constraint': keras.constraints.serialize(self.bias_constraint)
                  }

        base_config = super().get_config()
        return {**base_config, **config}

    def compute_output_shape(self, input_shape):
        output_shape = input_shape[0][0], self.output_dim
        return output_shape  # (batch_n_nodes, units) or (batch_n_nodes, units * 2)


class GCNAggregator(Layer):
    """
        Basic graphSAGE convolution layer as in: 
        [Inductive Representation Learning on Large Graphs](https://arxiv.org/abs/1706.02216)
        Tensorflow 1.x implementation: https://github.com/williamleif/GraphSAGE
        Pytorch implementation: https://github.com/williamleif/graphsage-simple/

        Aggregates via mean followed by matmul and non-linearity.
        Same matmul parameters are used self vector and neighbor vectors.    

        `GCNAggregator` implements the operation:
        `output = activation(Agg(Concat(neigh_x, x)) @ kernel) + bias)`
        where `x` is the feature matrix, `neigh_x` is the feature matrix of neighbors,
        `Agg` is the operation of aggregation (`mean`, `sum`, `max`, `min`) along the last dimension,
        `activation` is the element-wise activation function
        passed as the `activation` argument, `kernel` is a weights matrix
        created by the layer, and `bias` is a bias vector created by the layer
        (only applicable if `use_bias` is `True`).


        Arguments:
          units: Positive integer, dimensionality of the output space.
          use_bias: Boolean, whether the layer uses a bias vector.
          agg_method: String, which method of aggregation will be used for neighbor feature matrix,
            one of (`mean`, `sum`, `max`, `min`) will be used.
          activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).            
          kernel_initializer: Initializer for the `kernel` weights matrix.
          bias_initializer: Initializer for the bias vector.
          kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
          bias_regularizer: Regularizer function applied to the bias vector.
          activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
          kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
          bias_constraint: Constraint function applied to the bias vector.

        Input shape:
          tuple/list with two tensor: 2-D Tensor `x` and 3-D Tensor `neigh_x`: 
          `[(batch_n_nodes, n_features), (batch_n_nodes, n_samples, n_features)]`.
          The former one is the node feature matrix (Tensor) and the last is the neighbor feature matrix (Tensor).

        Output shape:
          2-D tensor with shape: `(batch_n_nodes, units)` or `(batch_n_nodes, units * 2)`,
          depend on using `Add` or `Concat` for the node and neighbor features.       
    """

    def __init__(self,
                 units,
                 use_bias=True,
                 agg_method='mean',
                 activation=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):

        kwargs.pop('concat', None)  # in order to be compatible with `MeanAggregator`
        super().__init__(**kwargs)
        self.units = units
        self.use_bias = use_bias
        self.agg_method = agg_method
        self.aggregator = {'mean': tf.reduce_mean, 'sum': tf.reduce_sum,
                           'max': tf.reduce_max, 'min': tf.reduce_min}[agg_method]
        self.activation = activations.get(activation)

        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)

        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)

        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

    def build(self, input_shape):
        input_dim = input_shape[0][-1]

        self.kernel = self.add_weight(shape=(input_dim, self.units),
                                      initializer=self.kernel_initializer,
                                      regularizer=self.kernel_regularizer,
                                      constraint=self.kernel_constraint,
                                      name='kernel')

        # Layer bias
        if self.use_bias:
            self.bias = self.add_weight(shape=(self.units, ),
                                        initializer=self.bias_initializer,
                                        regularizer=self.bias_regularizer,
                                        constraint=self.bias_constraint,
                                        name='bias')

        super().build(input_shape)

    def call(self, inputs):

        x, neigh_x = inputs
        x = tf.expand_dims(x, axis=1)
        agg = tf.concat([x, neigh_x], axis=1)
        h = self.aggregator(agg, axis=1)

        output = h @ self.kernel

        if self.use_bias:
            output += self.bias

        return self.activation(output)

    def get_config(self):

        config = {'units': self.units,
                  'use_bias': self.use_bias,
                  'agg_method': self.agg_method,
                  'activation': keras.activations.serialize(self.activation),
                  'kernel_initializer': keras.initializers.serialize(
                      self.kernel_initializer),
                  'bias_initializer': keras.initializers.serialize(
                      self.bias_initializer),
                  'kernel_regularizer': keras.regularizers.serialize(
                      self.kernel_regularizer),
                  'bias_regularizer': keras.regularizers.serialize(
                      self.bias_regularizer),
                  'activity_regularizer': keras.regularizers.serialize(
                      self.activity_regularizer),
                  'kernel_constraint': keras.constraints.serialize(
                      self.kernel_constraint),
                  'bias_constraint': keras.constraints.serialize(self.bias_constraint)
                  }

        base_config = super().get_config()
        return {**base_config, **config}

    def compute_output_shape(self, input_shape):
        output_shape = input_shape[0][0], self.units
        return tf.TensorShape(output_shape)  # (batch_n_nodes, units) or (batch_n_nodes, units * 2)
